from ._basedata import _BaseData
from openghg.plotting import plot_timeseries as general_plot_timeseries
import plotly.graph_objects as go
from typing import Any, Iterator, Optional


class ObsData(_BaseData):
    """This class is used to return observations data. It be created with a preloaded xarray Dataset or
    with a UUID and version number to retrieve data from Datasource zarr store.
    """

    # Compatability layer for legacy format - mimicking the behaviour of a dictionary
    # Previous format expected a dictionary containing the site code and data
    # as key:value pairs.
    # TODO: May also want to check other expected keys within the legacy
    # dictionary format and add them below
    def __getitem__(self, key: str) -> Any:
        """
        Returns the data attribute (xarray Dataset) when the site name is
        specified.
        Included as a compatability layer for legacy format as a dictionary
        containing a Dataset for each site code.

        key (str): Site code
        """
        site = self.metadata["site"].lower()
        if key.lower() == site:
            return self.data
        else:
            raise KeyError(f"Site '{key}' does not match to expected site '{site}'")

    def __iter__(self) -> Iterator:
        """
        Returns site code as the key for the dictionary as would be expected.
        """
        site = self.metadata["site"]
        return iter([site])

    def __len__(self) -> int:
        """
        Returns number of key values (fixed at 1 at present)
        """
        # Fixed length as 1 at the moment but may need to update if other key
        # values are added.
        return 1

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObsData):
            return NotImplemented

        if self.data is None or other.data is None:
            raise ValueError("Cannot compare data if it is not loaded")

        return self.data.equals(other.data) and self.metadata == other.metadata

    def plot_timeseries(
        self,
        title: Optional[str] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        units: Optional[str] = None,
        logo: Optional[bool] = True,
    ) -> go.Figure:
        """Plot a timeseries"""

        return general_plot_timeseries(
            data=self,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            units=units,
            logo=logo,
        )



from __future__ import annotations
import logging
import inspect
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Optional, Tuple, Union
import warnings
import numpy as np
from numpy import ndarray
from openghg.store import DataSchema
from openghg.store.base import BaseStore
from openghg.util import synonyms

from xarray import DataArray, Dataset

__all__ = ["Flux"]


logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


ArrayType = Optional[Union[ndarray, DataArray]]

logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class Flux(BaseStore):
    """This class is used to process flux / emissions flux data"""

    _data_type = "flux"
    _root = "Flux"
    _uuid = "c5c88168-0498-40ac-9ad3-949e91a30872"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    def read_data(self, binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Optional[Dict]:
        """Ready a footprint from binary data

        Args:
            binary_data: Footprint data
            metadata: Dictionary of metadata
            file_metadat: File metadata
        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        with TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)

            try:
                filename = file_metadata["filename"]
            except KeyError:
                raise KeyError("We require a filename key for metadata read.")

            filepath = tmpdir_path.joinpath(filename)
            filepath.write_bytes(binary_data)

            return self.read_file(filepath=filepath, **metadata)

    def read_file(
        self,
        filepath: Union[str, Path],
        species: str,
        source: str,
        domain: str,
        database: Optional[str] = None,
        database_version: Optional[str] = None,
        model: Optional[str] = None,
        source_format: str = "openghg",
        time_resolved: bool = False,
        high_time_resolution: bool = False,
        period: Optional[Union[str, tuple]] = None,
        chunks: Optional[Dict] = None,
        continuous: bool = True,
        if_exists: str = "auto",
        save_current: str = "auto",
        overwrite: bool = False,
        force: bool = False,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        optional_metadata: Optional[Dict] = None,
    ) -> dict:
        """Read flux / emissions file

        Args:
            filepath: Path of flux / emissions file
            species: Species name
            domain: Flux / Emissions domain
            source: Flux / Emissions source
            database: Name of database source for this input (if relevant)
            database_version: Name of database version (if relevant)
            model: Model name (if relevant)
            source_format : Type of data being input e.g. openghg (internal format)
            time_resolved: If this is a high resolution file
            high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
            period: Period of measurements. Only needed if this can not be inferred from the time coords
            If specified, should be one of:
                - "yearly", "monthly"
                - suitable pandas Offset Alias
                - tuple of (value, unit) as would be passed to pandas.Timedelta function
            chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
                for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
                See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
                To disable chunking pass in an empty dictionary.
            continuous: Whether time stamps have to be continuous.
            if_exists: What to do if existing data is present.
                - "auto" - checks new and current data for timeseries overlap
                   - adds data if no overlap
                   - raises DataOverlapError if there is an overlap
                - "new" - just include new data and ignore previous
                - "combine" - replace and insert new data into current timeseries
            save_current: Whether to save data in current form and create a new version.
                - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
                - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
                - "n" / "no" - Allow current data to updated / deleted
            overwrite: Deprecated. This will use options for if_exists="new".
            force: Force adding of data even if this is identical to data stored.
            compressor: A custom compressor to use. If None, this will default to
                `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
                See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
            filters: Filters to apply to the data on storage, this defaults to no filtering. See
                https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
            optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
        Returns:
            dict: Dictionary of datasource UUIDs data assigned to
        """
        from openghg.types import FluxTypes
        from openghg.util import (
            clean_string,
            load_flux_parser,
            check_if_need_new_version,
        )

        species = clean_string(species)
        species = synonyms(species)
        source = clean_string(source)
        domain = clean_string(domain)

        if high_time_resolution:
            warnings.warn(
                "This argument is deprecated and will be replaced in future versions with time_resolved.",
                DeprecationWarning,
            )
            time_resolved = high_time_resolution

        if overwrite and if_exists == "auto":
            logger.warning(
                "Overwrite flag is deprecated in preference to `if_exists` (and `save_current`) inputs."
                "See documentation for details of these inputs and options."
            )
            if_exists = "new"

        # Making sure new version will be created by default if force keyword is included.
        if force and if_exists == "auto":
            if_exists = "new"

        new_version = check_if_need_new_version(if_exists, save_current)

        filepath = Path(filepath)

        try:
            source_format = FluxTypes[source_format.upper()].value
        except KeyError:
            raise ValueError(f"Unknown data type {source_format} selected.")

        # Load the data retrieve object
        parser_fn = load_flux_parser(source_format=source_format)

        _, unseen_hashes = self.check_hashes(filepaths=filepath, force=force)

        if not unseen_hashes:
            return {}

        filepath = next(iter(unseen_hashes.values()))

        if chunks is None:
            chunks = {}

        # Define parameters to pass to the parser function
        # TODO: Update this to match against inputs for parser function.
        param = {
            "filepath": filepath,
            "species": species,
            "domain": domain,
            "source": source,
            "time_resolved": time_resolved,
            "period": period,
            "continuous": continuous,
            "data_type": "flux",
            "chunks": chunks,
        }

        optional_keywords: dict[Any, Any] = {
            "database": database,
            "database_version": database_version,
            "model": model,
        }

        signature = inspect.signature(parser_fn)
        fn_accepted_parameters = [param.name for param in signature.parameters.values()]

        input_parameters: dict[Any, Any] = param.copy()

        # Checks if optional parameters are present in function call and includes them else ignores its inclusion in input_parameters.
        for param, param_value in optional_keywords.items():
            if param in fn_accepted_parameters:
                input_parameters[param] = param_value
            else:
                logger.warning(
                    f"Input: '{param}' (value: {param_value}) is not being used as part of the standardisation process."
                    f"This is not accepted by the current standardisation function: {parser_fn}"
                )

        flux_data = parser_fn(**input_parameters)

        # Checking against expected format for Flux
        for split_data in flux_data.values():
            em_data = split_data["data"]
            Flux.validate_data(em_data)

        min_required = ["species", "source", "domain"]
        for key, value in optional_keywords.items():
            if value is not None:
                min_required.append(key)

        required = tuple(min_required)

        if optional_metadata:
            common_keys = set(required) & set(optional_metadata.keys())

            if common_keys:
                raise ValueError(
                    f"The following optional metadata keys are already present in required keys: {', '.join(common_keys)}"
                )
            else:
                for key, parsed_data in flux_data.items():
                    parsed_data["metadata"].update(optional_metadata)

        data_type = "flux"
        datasource_uuids = self.assign_data(
            data=flux_data,
            if_exists=if_exists,
            new_version=new_version,
            data_type=data_type,
            required_keys=required,
            compressor=compressor,
            filters=filters,
        )

        # Record the file hash in case we see this file again
        self.store_hashes(unseen_hashes)

        return datasource_uuids

    def transform_data(
        self,
        datapath: Union[str, Path],
        database: str,
        if_exists: str = "auto",
        save_current: str = "auto",
        overwrite: bool = False,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        optional_metadata: Optional[Dict] = None,
        **kwargs: Dict,
    ) -> Dict:
        """
        Read and transform a flux / emissions database. This will find the appropriate
        parser function to use for the database specified. The necessary inputs
        are determined by which database is being used.

        The underlying parser functions will be of the form:
            - openghg.transform.flux.parse_{database.lower()}
                - e.g. openghg.transform.flux.parse_edgar()

        Args:
            datapath: Path to local copy of database archive (for now)
            database: Name of database
            if_exists: What to do if existing data is present.
                - "auto" - checks new and current data for timeseries overlap
                   - adds data if no overlap
                   - raises DataOverlapError if there is an overlap
                - "new" - just include new data and ignore previous
                - "combine" - replace and insert new data into current timeseries
            save_current: Whether to save data in current form and create a new version.
                - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
                - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
                - "n" / "no" - Allow current data to updated / deleted
            overwrite: Deprecated. This will use options for if_exists="new".
            compressor: A custom compressor to use. If None, this will default to
                `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
                See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
            filters: Filters to apply to the data on storage, this defaults to no filtering. See
                https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
            **kwargs: Inputs for underlying parser function for the database.

                Necessary inputs will depend on the database being parsed.

        TODO: Could allow Callable[..., Dataset] type for a pre-defined function be passed
        """
        import inspect
        from openghg.types import FluxDatabases
        from openghg.util import load_flux_database_parser, check_if_need_new_version

        if overwrite and if_exists == "auto":
            logger.warning(
                "Overwrite flag is deprecated in preference to `if_exists` (and `save_current`) inputs."
                "See documentation for details of these inputs and options."
            )
            if_exists = "new"

        new_version = check_if_need_new_version(if_exists, save_current)

        datapath = Path(datapath)

        try:
            data_type = FluxDatabases[database.upper()].value
        except KeyError:
            raise ValueError(f"Unable to transform '{database}' selected.")

        # Load the data retrieve object
        parser_fn = load_flux_database_parser(database=database)

        # Find all parameters that can be accepted by parse function
        all_param = list(inspect.signature(parser_fn).parameters.keys())

        # Define parameters to pass to the parser function from kwargs
        param: Dict[Any, Any] = {key: value for key, value in kwargs.items() if key in all_param}
        param["datapath"] = datapath  # Add datapath explicitly (for now)

        flux_data = parser_fn(**param)

        # Checking against expected format for Flux
        for split_data in flux_data.values():
            em_data = split_data["data"]
            Flux.validate_data(em_data)

        required_keys = ("species", "source", "domain")

        if optional_metadata:
            common_keys = set(required_keys) & set(optional_metadata.keys())

            if common_keys:
                raise ValueError(
                    f"The following optional metadata keys are already present in required keys: {', '.join(common_keys)}"
                )
            else:
                for key, parsed_data in flux_data.items():
                    parsed_data["metadata"].update(optional_metadata)

        data_type = "flux"
        datasource_uuids = self.assign_data(
            data=flux_data,
            if_exists=if_exists,
            new_version=new_version,
            data_type=data_type,
            required_keys=required_keys,
            compressor=compressor,
            filters=filters,
        )

        return datasource_uuids

    @staticmethod
    def schema() -> DataSchema:
        """
        Define schema for flux / emissions Dataset.

        Includes flux/emissions for each time and position:
            - "flux"
                - expected dimensions: ("time", "lat", "lon")

        Expected data types for all variables and coordinates also included.

        Returns:
            DataSchema : Contains schema for Flux.
        """
        data_vars: Dict[str, Tuple[str, ...]] = {"flux": ("time", "lat", "lon")}
        dtypes = {"lat": np.floating, "lon": np.floating, "time": np.datetime64, "flux": np.floating}

        data_format = DataSchema(data_vars=data_vars, dtypes=dtypes)

        return data_format

    @staticmethod
    def validate_data(data: Dataset) -> None:
        """
        Validate input data against Flux schema - definition from
        Flux.schema() method.

        Args:
            data : xarray Dataset in expected format

        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the Flux schema.
        """
        data_schema = Flux.schema()
        data_schema.validate_data(data)



"""
This is used as a base for the other dataclasses and shouldn't be used directly.
"""

from typing import Dict, Optional, Union
from openghg.store.storage import LocalZarrStore
import xarray as xr
from pandas import Timestamp, Timedelta
import logging

logger = logging.getLogger("openghg.dataobjects")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class _BaseData:
    def __init__(
        self,
        metadata: Dict,
        data: Optional[xr.Dataset] = None,
        uuid: Optional[str] = None,
        version: Optional[str] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        sort: bool = True,
        elevate_inlet: bool = False,
        attrs_to_check: Optional[Dict] = None,
    ) -> None:
        """
        This handles data for each of the data type classes. It accepts either a Dataset
        or a UUID and version to lazy load a Dataset from a zarr store. If dates are passed then the
        dataset is sliced to the requested time period.

        Args:
            metadata: Dictionary of metadata
            data: Dataset if data is already loaded
            uuid: UUID of Datasource to retrieve data from
            version: Version of data requested from Datasource
            start_date: Start date of data to retrieve
            end_date: End date of data to retrieve
            sort: Sort the resulting Dataset by the time dimension, defaults to False
            elevate_inlet: Force the elevation of the inlet attribute
            attrs_to_check: Attributes to check for duplicates. If duplicates are present
                a new data variable will be created containing the values from each dataset
                If a dictionary is passed, the attribute(s) will be retained and the new value assigned.
                If a list/string is passed, the attribute(s) will be removed.
        """
        from openghg.util import timestamp_epoch, timestamp_now

        if data is None and uuid is None and version is None:
            raise ValueError("Must supply either data or uuid and version")

        self.metadata = metadata
        self._uuid = uuid

        self._start_date = start_date
        self._end_date = end_date
        self._zarrstore = None

        if elevate_inlet:
            raise NotImplementedError("elevate_inlet not implemented yet")

        if attrs_to_check is not None:
            raise NotImplementedError("attrs_to_check not implemented yet")

        sorted = False  # Check so we don't sort more than once

        if data is not None:
            self.data = data
        elif uuid is not None and version is not None:
            slice_time = False
            if start_date is not None or end_date is not None:
                slice_time = True
                if start_date is None:
                    start_date = timestamp_epoch()
                if end_date is None:
                    end_date = timestamp_now()

            self._version = version
            self._bucket = metadata["object_store"]

            self._zarrstore = LocalZarrStore(bucket=self._bucket, datasource_uuid=uuid, mode="r")

            self.data = self._zarrstore.get(version=version)
            if slice_time:
                # If slicing by time, this must be sorted along the time dimension
                self.data = self.data.sortby("time")
                sorted = True

                if self.data.time.size > 1:
                    start_date = start_date - Timedelta("1s")
                    # TODO: May want to consider this extra 1s subtraction as end_date on data has already has -1s applied.
                    end_date = end_date - Timedelta("1s")

                    # TODO - I feel we should do this in a tider way
                    start_date = start_date.tz_localize(None)
                    end_date = end_date.tz_localize(None)

                    self.data = self.data.sel(time=slice(start_date, end_date))
        else:
            raise ValueError(
                "Must supply either data or uuid and version, cannot create an empty data object."
            )

        if sort and not sorted:
            try:
                self.data = self.data.sortby("time")
            except KeyError:
                logger.debug("Cannot sort data by time as no time dimension present")

    def __bool__(self) -> bool:
        return bool(self.data)

    def __str__(self) -> str:
        return f"Data: {self.data}\nMetadata: {self.metadata}"



import json
import logging
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, Iterable

from openghg.dataobjects import ObsData
from openghg.util import running_on_hub
from pandas import DataFrame

__all__ = ["SearchResults"]

T = TypeVar("T", bound="SearchResults")

logger = logging.getLogger("openghg.dataobjects")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class SearchResults:
    """This class is used to return data from the search function. It
    has member functions to retrieve data from the object store.

    Args:
        keys: Dictionary of keys keyed by Datasource UUID
        metadata: Dictionary of metadata keyed by Datasource UUID
        start_result: ?
    """

    # TODO - WIP move to tinydb metadata lookup to simplify code
    def __init__(
        self,
        metadata: Optional[Dict] = None,
        start_result: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ):
        # db = tinydb.TinyDB(tinydb.storages.MemoryStorage)
        if metadata is not None:
            self.metadata = metadata
            # db.insert_multiple([m for m in metadata.values()])
            self.results = (
                DataFrame.from_dict(data=metadata, orient="index").reset_index().drop(columns="index")
            )

            if start_result is not None:
                for uuid_key, uuid_metadata in metadata.items():
                    if start_result in uuid_metadata:
                        other_keys = list(uuid_metadata.keys())
                        other_keys.remove(start_result)
                        reorder = [start_result] + other_keys
                        metadata[uuid_key] = {key: uuid_metadata[key] for key in reorder}

        else:
            self.results = {}  # type: ignore
            self.metadata = {}

        self._start_date = start_date
        self._end_date = end_date

        self.hub = running_on_hub()

    def __str__(self) -> str:
        SearchResults.df_to_table_console_output(df=DataFrame.from_dict(data=self.metadata))

        return f"Found {len(self.results)} results.\nView the results DataFrame using the results property."

    def __repr__(self) -> str:
        return self.__str__()

    def __bool__(self) -> bool:
        return bool(self.metadata)

    def __len__(self) -> int:
        return len(self.metadata)

    # def __iter__(self) -> Iterator:
    #     yield from self.results.iterrows()

    def to_data(self) -> Dict:
        """Convert this object to a dictionary for JSON serialisation

        Returns:
            dict: Dictionary of data
        """
        return {
            "metadata": self.metadata,
            "hub": self.hub,
        }

    def to_json(self) -> str:
        """Serialises the object to JSON

        Returns:
            str: JSON str
        """
        return json.dumps(self.to_data())

    @classmethod
    def from_json(cls: Type[T], data: Union[bytes, str]) -> T:
        """Create a SearchResults object from a dictionary

        Args:
            data: Serialised object
        Returns:
            SearchResults: SearchResults object
        """
        loaded = json.loads(data)

        return cls(metadata=loaded["metadata"])

    def retrieve(
        self,
        dataframe: Optional[DataFrame] = None,
        version: str = "latest",
        **kwargs: Any,
    ) -> Union[ObsData, List[ObsData]]:
        """Retrieve data from object store using a filtered pandas DataFrame

        Args:
            dataframe: pandas DataFrame
            sort: Sort data by date in retrieved Dataset
            elevate_inlet: Elevate inlet to a variable within the Dataset, useful
            for ranked data.
        Returns:
            ObsData / List[ObsData]: ObsData object(s)
        """
        if dataframe is not None:
            uuids = dataframe["uuid"].to_list()
            return self._retrieve_by_uuid(uuids=uuids, version=version)
        else:
            return self._retrieve_by_term(version=version, **kwargs)

    def retrieve_all(
        self,
        version: str = "latest",
    ) -> Union[ObsData, List[ObsData]]:
        """Retrieves all data found during the search

        Args:
            sort: Sort by time. Note that this may be very memory hungry for large Datasets.
            elevate_inlet: Elevate inlet to a variable within the Dataset, useful
            for ranked data.
        Returns:
            ObsData / List[ObsData]: ObsData object(s)
        """
        return self._retrieve_by_uuid(uuids=self.metadata.keys(), version=version)

    def uuids(self) -> List:
        """Return the UUIDs of the found data

        Returns:
            list: List of UUIDs
        """
        return list(self.metadata.keys())

    def _retrieve_by_term(self, version: str, **kwargs: Any) -> Union[ObsData, List[ObsData]]:
        """Retrieve data from the object store by search term. This function scans the
        metadata of the retrieved results, retrieves the UUID associated with that data,
        pulls it from the object store, recombines it into an xarray Dataset and returns
        ObsData object(s).

        Args:
            kwargs: Metadata values to search for
        """
        uuids = set()
        # Make sure we don't have any Nones
        clean_kwargs = {k: v for k, v in kwargs.items() if v is not None}

        for uid, metadata in self.metadata.items():
            n_required = len(clean_kwargs)
            n_matched = 0
            for key, value in clean_kwargs.items():
                try:
                    # Here we want to check if it's a list and if so iterate over it
                    if isinstance(value, (list, tuple)):
                        for val in value:
                            val = str(val).lower()
                            if metadata[key.lower()] == val:
                                n_matched += 1
                                break
                    else:
                        value = str(value).lower()
                        if metadata[key.lower()] == value:
                            n_matched += 1
                except KeyError:
                    pass

            if n_matched == n_required:
                uuids.add(uid)

        # Now we can retrieve the data using the UUIDs
        return self._retrieve_by_uuid(uuids=list(uuids), version=version)

    def _retrieve_by_uuid(self, uuids: Iterable, version: str) -> Union[ObsData, List[ObsData]]:
        """Internal retrieval function that uses the passed in UUIDs to retrieve
        the keys from the key_data dictionary, pull the data from the object store,
        create ObsData object(s) and return the result.

        Args:
            uuids: UUIDs of Datasources in the object store
        Returns:
            ObsData / List[ObsData]: ObsData object(s)
        """
        results = []
        for uuid in uuids:
            metadata = self.metadata[uuid]
            if version == "latest":
                version = metadata["latest_version"]
            else:
                if version not in metadata["versions"]:
                    raise ValueError(f"Invalid version {version} for UUID {uuid}")

            results.append(
                ObsData(
                    uuid=uuid,
                    version=version,
                    metadata=metadata,
                    start_date=self._start_date,
                    end_date=self._end_date,
                )
            )

        if len(results) == 1:
            return results[0]
        else:
            return results

    @staticmethod
    def df_to_table_console_output(df: DataFrame) -> None:
        """
        Process the DataFrame and display it as a formatted table in the console.

        Args:
            df (DataFrame): The DataFrame to be processed and displayed.

        Returns:
            None
        """
        try:
            from rich import print
            from rich.table import Table, box
        except ModuleNotFoundError:
            logger.warning("Unable to use rich package to display search results. Please install rich")
            return None

        # Split columns into sets
        column_sets = [df.columns[i : i + 4] for i in range(0, len(df.columns), 4)]

        # Iterate over the column sets
        for columns in column_sets:
            # Create a table instance
            table = Table(show_header=False, header_style="bold", box=box.HORIZONTALS)

            # Add table headers
            for i, column in enumerate(columns, start=1):
                if i == 1:
                    table.add_column(column, style="bold")
                else:
                    table.add_column(column)

            # Add table data
            for index, row in df.iterrows():
                row_data = [str(index)] + [str(row[column]) for column in columns]
                table.add_row(*row_data)

            # Print the table
            print(table)
            print()



"""
This module implements a MetaStore based on TinyDB, with custom
middleware based on `CachingMiddleware`.

The metastore can be accessed given a bucket and data type.
 In a context manager, use `open_metastore`. Outside of a context
 manager, use `ClassicMetaStore.from_bucket`; in this case, the
metastore must be closed using the `close` method.

Closing the metastore is necessary, since `CachingMiddleware`
doesn't write to disk unless at least 1000 writes have been made.

The `ClassicMetaStore` is organised in the same way that the
metastore was organised in OpenGHG <= v 6.2: there are separate
TinyDB databases for each data type.

Besides taking a data type parameter in the `__init__` method,
`ClassicMetaStore` only extends `TinyDBMetastore` by adding
a `from_bucket` class method, and a `close` method.
These methods are only needed by `store.BaseStore`.

"""

from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
import json
from typing import Any, cast, Literal, Optional, Type

from filelock import FileLock
from openghg.objectstore import exists, get_object, set_object_from_json, get_object_lock_path
from openghg.objectstore.metastore import TinyDBMetaStore
from openghg.types import MetastoreError
from openghg.util import hash_string
import tinydb
from tinydb.middlewares import Middleware


def get_metakey(data_type: str) -> str:
    """Return the metakey for a given data type.

    Args:
        data_type: data type to get metakey for.

    Returns:
        Metakey string for given data type, if found, or "default"
            if data type not found.
    """
    from openghg.store import data_class_info

    object_store_data_classes = data_class_info()
    try:
        result = object_store_data_classes[data_type]
    except KeyError:
        return "default"
    return f"{result['_root']}/uuid/{result['_uuid']}/metastore"


class BucketKeyStorage(tinydb.Storage):
    """Custom TinyDB storage class.

    Uses methods in `_local_store` module to read/write files via bucket and key.
    """

    def __init__(self, bucket: str, key: str, mode: Literal["r", "rw"]) -> None:
        """Create BucketKeyStorage object.

        Args:
            bucket: path to object store bucket (as string)
            key: metastore key
            mode: "r" for read-only, "rw" for read/write

        Returns:
            None
        """
        valid_modes = ("r", "rw")
        if mode not in valid_modes:
            raise ValueError(f"Invalid mode, please choose one of {valid_modes}.")

        self._key = key
        self._bucket = bucket
        self._mode = mode

    def read(self) -> Optional[dict]:
        """Read data from database.

        Returns:
            Dictionary version of JSON database, or None if database has
            not been initialised. (Returning None is required by the TinyDB
            interface.)
        """
        key = self._key

        if not exists(bucket=self._bucket, key=key):
            return None

        data = get_object(bucket=self._bucket, key=self._key)

        try:
            json_data: dict = json.loads(data)
            return json_data
        except json.JSONDecodeError:
            return None

    def write(self, data: dict) -> None:
        """Write data to metastore.

        Args:
            data: dictonary of data to add

        Returns:
            None

        Raises:
            MetastoreError if metastore opened in read-only mode.
        """
        if self._mode == "r":
            raise MetastoreError("Cannot write to metastore in read-only mode.")

        key = self._key
        set_object_from_json(bucket=self._bucket, key=key, data=data)

    def close(self) -> None:
        pass


class SafetyCachingMiddleware(Middleware):
    """Middleware that caches changes to the database, and writes
    these changes when the database is closed. Changes are only written
    if the underlying file has not changed. (The underlying file is the
    persistent record of the database that is read in by the storage class.)

    This differs from CachingMiddleware in two ways:
    1) changes are never written before closing the database, whereas
    CachingMiddleware will write after 1000 changes by default.

    2) CachingMiddleware does not check if the underlying file
    has changed since it was first accessed by the storage class.
    """

    def __init__(self, storage_cls: Type[tinydb.Storage]) -> None:
        """Follows the standard pattern for middleware.

        Args:
            storage_cls: tinydb storage class to wrap with Middleware
        Returns:
            None
        """
        super().__init__(storage_cls)
        self.cache = None  # in-memory version of database
        self.database_hash = None  # hash taken when database first read
        self.writes_made = False  # flag to check if writes made

    def read(self):
        """Read the database from the cache, if present, otherwise load
        the database from the underlying storage and save a hash of the result.
        """
        if self.cache is None:
            self.cache = self.storage.read()
            self.database_hash = hash_string(str(self.cache))

        return self.cache

    def write(self, data):
        """Store data in the cache.

        Args:
            data: data to store (this is used internally by TinyDB)
        """
        self.cache = data
        self.writes_made = True

    def close(self):
        """Close the database. If writes have been made, and the underlying
        file has not changed, writes will be saved to disk at this point.

        Raises: MetaStoreError if writes have been made and the underlying file *has* been changed.
        """
        if self.writes_made:
            # we know that the cache is a dictionary of dictionaries
            self.cache = cast(dict[str, dict[str, Any]], self.cache)

            # check if stored hash matches current hash
            if self.database_hash == hash_string(str(self.storage.read())):
                # if underlying file not changed, write data
                self.storage.write(self.cache)
            else:
                raise MetastoreError(
                    "Could not write to object store: object store modified while write in progress."
                )

        # if close is called explicitly, rather than through a context manager,
        # then the cache should be empty, otherwise if the metastore instance is reused
        # it won't reflect the actual state of the metastore. (This is an edge case.)
        self.cache = None

        # let underlying storage clean up
        self.storage.close()


@contextmanager
def open_metastore(
    bucket: str, data_type: str, mode: Literal["r", "rw"] = "rw"
) -> Generator[TinyDBMetaStore, None, None]:
    """Context manager for TinyDBMetaStore based on OpenGHG v<=6.2 set-up for keys
    and TinyDB.

    Args:
        bucket: path to object store
        data_type: data type of metastore to open
        mode: 'rw' for read/write, 'r' for read only

    Yields:
        ClassicMetaStore instance.
    """
    key = get_metakey(data_type)
    with tinydb.TinyDB(bucket, key, mode, storage=SafetyCachingMiddleware(BucketKeyStorage)) as db:
        metastore = TinyDBMetaStore(database=db)
        yield metastore


class DataClassMetaStore(TinyDBMetaStore):
    """Class that allows:
    - creating a TinyDB MetaStore via bucket and data type
    - locking the MetaStore

    There is also a `close` method, for use inside the `BaseStore` context-manager.
    """

    def __init__(self, bucket: str, data_type: str) -> None:
        self.data_type = data_type
        self.key = get_metakey(data_type)
        database = tinydb.TinyDB(
            bucket, self.key, mode="rw", storage=SafetyCachingMiddleware(BucketKeyStorage)
        )
        super().__init__(database=database)

        lock_path = get_object_lock_path(bucket, self.key)
        self.lock = FileLock(lock_path, timeout=600)  # file lock with 10 minute timeout

    def acquire_lock(self) -> None:
        """Acquire a lock for the object store."""
        self.lock.acquire(poll_interval=1)

    def release_lock(self) -> None:
        """Acquire a lock for the object store."""
        self.lock.release()

    def close(self) -> None:
        """Close the underlying TinyDB database."""
        self.lock.release()
        self._db.close()



from __future__ import annotations
import logging
from typing import Any, Dict, Literal, Iterator, Union, Optional, MutableMapping
import xarray as xr
import zarr
import os
import re
import shutil

from openghg.types import ZarrStoreError
from pathlib import Path
from openghg.objectstore import get_folder_size
from openghg.store.storage import Store

logger = logging.getLogger("openghg.store.base")
logger.setLevel(logging.DEBUG)

StoreLike = Union[zarr.storage.BaseStore, MutableMapping]


class LocalZarrStore(Store):
    """A zarr based data store on the local filesystem.
    This is used by Datasource to handle the storage of versioned data.

    Args:
        bucket: path to object store bucket
        datasource_uuid: Datasource UUID
        mode: Opening mode for store, "rw" for read-write, "r" for read-only
    """

    def __init__(self, bucket: str, datasource_uuid: str, mode: Literal["rw", "r"] = "rw") -> None:
        self._bucket = bucket
        self._mode = mode
        self._root_store_key = f"data/{datasource_uuid}/zarr"
        self._stores_path = Path(bucket, self._root_store_key).expanduser().resolve()

        # Here we ensure we have the correct directory structure for the zarr stores
        # and do a lookup for existing stores, populating the _stores dictionary.
        self._stores = {}
        if not self._stores_path.exists():
            self._stores_path.mkdir(parents=True)
        else:
            compiled_reg = re.compile(r"v\d+")
            for f in sorted(os.listdir(self._stores_path)):
                if compiled_reg.match(str(f)):
                    full_path = Path(self._stores_path, f).expanduser().resolve()
                    self._stores[f] = zarr.storage.DirectoryStore(full_path)
        # An in memory store used if data is popped from the store
        self._memory_store: Dict[str, bytes] = {}

    def __bool__(self) -> bool:
        return any(self._stores)

    def _check_version(self, version: str) -> str:
        """Check if the given version exists in the store.
        Raises a ZarrStoreError if the version does not exist.

        Args:
            version: Data version
        Returns:
            str: Lowercase version
        """
        if version.lower() not in self._stores:
            raise ZarrStoreError(f"Invalid version - {version}")

        return version.lower()

    def _check_writable(self) -> None:
        """Check if the store is writable

        Returns:
            None
        """
        if self._mode == "r":
            raise PermissionError("Cannot modify a read-only zarr store")

    def keys(self, version: str) -> Iterator[str]:
        """Keys of data stored in the zarr store.

        Args:
            version: Data version
        Returns:
            Generator: Generator object
        """
        version = self._check_version(version)
        keys: Iterator = self._stores[version].keys()
        return keys

    def close(self) -> None:
        """Close the zarr store.

        Returns:
            None
        """
        for store in self._stores.values():
            store.close()

    def store_key(self, version: str) -> str:
        """Return the key of this zarr Store

        Args:
            version: Data version
        Returns:
            str: Key of zarr store
        """
        if version.lower() not in self._stores:
            raise KeyError(f"Invalid version - {version}")

        return str(Path(self._root_store_key, version))

    def store_path(self, version: str) -> Path:
        """Return the path of this zarr Store

        Args:
            version: Data version
        Returns:
            Path: Path of zarr store
        """
        return Path(self._stores_path, version)

    def version_exists(self, version: str) -> bool:
        """Check if a version exists in the current store

        Args:
            version: Data version
        Returns:
            bool: True if version exists
        """
        return version.lower() in self._stores

    def add(
        self,
        version: str,
        dataset: xr.Dataset,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        append_dim: str = "time",
    ) -> None:
        """Add an xr.Dataset to the zarr store.

        Args:
            version: Data version
            dataset: xr.Dataset to add
            compressor: Compression for zarr encoding
            filters: Filters for zarr encoding
            append_dim: Dimension to append to
        Returns:
            None
        """
        from openghg.store.storage import get_zarr_encoding

        self._check_writable()
        version = version.lower()

        # Append new data to the zarr store for the current version
        if version in self._stores:
            # We want to ensure the chunking of the incoming data matches the chunking of the stored data
            chunking = self.match_chunking(version=version, dataset=dataset)
            if chunking:
                dataset = dataset.chunk(chunking)

            dataset.to_zarr(
                store=self._stores[version], mode="a", consolidated=True, append_dim=append_dim, compute=True
            )
        # Otherwise we create a new zarr Store for the version
        else:
            if not self._stores and version != "v1":
                raise ValueError("First version must be v1")

            self._stores[version] = zarr.storage.DirectoryStore(self.store_path(version=version))
            encoding = get_zarr_encoding(data_vars=dataset.data_vars, filters=filters, compressor=compressor)
            dataset.to_zarr(
                store=self._stores[version], mode="w", encoding=encoding, consolidated=True, compute=True
            )

    def get(self, version: str) -> xr.Dataset:
        """Get the version of the dataset stored in the zarr store.

        Args:
            version: Data version
        Returns:
            xr.Dataset: Dataset from the store
        """
        version = self._check_version(version)
        store = self._stores[version]
        ds: xr.Dataset = xr.open_zarr(store=store, consolidated=True)
        return ds

    def _pop(self, version: str) -> xr.Dataset:
        """Pop some data from the store. This copies the data in the version specified
        to a memory store, deletes the version and returns an xarray Dataset loaded from the
        memory store.

        Note that this will delete the data from the store on the local filesystem.

        Args:
            version: Data version
        Returns:
            Dataset: Dataset popped from the store
        """
        raise NotImplementedError("This method will be updated to ensure data is backed up.")
        self._check_writable()
        version = self._check_version(version)

        self._memory_store.clear()
        store = self._stores[version]
        # Let's copy the data we want to pop into a memory store and return it from there
        zarr.convenience.copy_store(source=store, dest=self._memory_store)
        self.delete_version(version=version)
        ds: xr.Dataset = xr.open_zarr(store=self._memory_store)
        return ds

    def _copy_to_memorystore(self, version: str) -> Dict:
        """Copies the compressed data from the filesystem store to an in-memory store.
        This preserves the compression and chunking of the data and the store
        can be opened as a single dataset. This may be useful for testing

        Args:
            version: Data version
        Returns:
            dict: In-memory copy of compressed data
        """
        version = self._check_version(version)
        store = self._stores[version]
        mem_store: Dict[str, bytes] = {}
        zarr.copy_store(source=store, dest=mem_store)
        return mem_store

    def delete_version(self, version: str) -> None:
        """Delete a version from the store

        Args:
            version: Data version
        Returns:
            None
        """
        self._check_writable()
        version = self._check_version(version)

        path = self.store_path(version=version)
        shutil.rmtree(path)
        del self._stores[version]

    def delete_all(self) -> None:
        """Delete all data from the zarr store.

        Returns:
            None
        """
        self._check_writable()
        self._stores.clear()
        shutil.rmtree(self._stores_path, ignore_errors=True)

    def update(
        self,
        version: str,
        dataset: xr.Dataset,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
    ) -> None:
        """Update a version of the data. This first deletes the current version
        and then adds the new version. To update a version in place, keeping the current data,
        use the append function.

        Args:
            version: Data version
            dataset: Dataset to add
            compressor: Compression for zarr encoding
            filters: Filters for zarr encoding
        Returns:
            None
        """
        self._check_writable()
        self.delete_version(version=version)
        self.add(version=version, dataset=dataset, compressor=compressor, filters=filters)

    def bytes_stored(self) -> int:
        """Return the number of bytes stored in the zarr store

        Returns:
            int: Number of bytes stored in zarr store
        """
        bytes_stored = 0
        for version in self._stores:
            bytes_stored += get_folder_size(folder_path=self.store_path(version=version))

        return bytes_stored

    def match_chunking(self, version: str, dataset: xr.Dataset) -> Dict[str, int]:
        """Ensure that chunks of incoming data match the chunking of the stored data.

        If no chunking is found then an empty dictionary is returned.
        If there is no mismatch then an empty dictionary is returned.
        Returns the chunking scheme of the stored data if there is a mismatch.

        Args:
            version: Data version
            dataset: Incoming dataset
        Returns:
            dict: Chunking scheme
        """
        version = self._check_version(version)

        incoming_chunks = dict(dataset.chunks)
        incoming_actually_chunked = {k: max(v) for k, v in incoming_chunks.items() if len(v) > 1}

        stored_chunks = {str(k): v for k, v in self.get(version=version).chunksizes.items()}
        # Only take chunks that aren't whole dimensions
        stored_actually_chunked = {k: max(v) for k, v in stored_chunks.items() if len(v) > 1}

        if not incoming_actually_chunked and not stored_actually_chunked:
            return {}

        if incoming_actually_chunked != stored_actually_chunked:
            if not incoming_actually_chunked:
                msg = (
                    f"Incoming data is not chunked, using stored chunking schema: {stored_actually_chunked}."
                )
            else:
                msg = (
                    f"Chunking scheme of incoming data ({incoming_actually_chunked}) does not match stored data."
                    + f"\nUsing stored chunking schema: {stored_actually_chunked}."
                    + "\nThis may result in an increased processing time. Not passing the chunks argument may be faster."
                )

            logger.warning(msg)

            # For coordinates we haven't chunked over we'll use the full size
            for k in dataset.dims:
                k = str(k)  # xr.Dataset.dims returns a Mapping with Hashable keys, which may not be strings
                if k not in stored_actually_chunked:
                    stored_actually_chunked[k] = dataset.sizes[k]

            return stored_actually_chunked

        return {}

    def match_attributes(self, version: str, dataset: xr.Dataset) -> Dict:
        """Ensure the attributes of the stored and incoming data are matched,
        any attributes that differ will be added as a new key with a number appended.
        e.g. author_1 for a differing author value.

        Args:
            version: Data version
            dataset: Incoming dataset
        Returns:
            dict: Dictionary of compared and updated attributes
        """
        raise NotImplementedError()



from typing import Any, Tuple, Dict, List

__all__ = ["define_data_types", "define_data_type_classes"]


def define_data_types() -> Tuple[str, ...]:
    """
    Define names of data types for objects within OpenGHG
    """

    data_types = (
        "surface",
        "column",
        "flux",
        # "met",
        "footprints",
        "boundary_conditions",
        "eulerian_model",
    )

    return data_types


def define_data_type_classes() -> Dict[str, Any]:
    """
    Define mapping between data types and associated input classes within OpenGHG
    """
    from openghg.store import (
        BoundaryConditions,
        Flux,
        EulerianModel,
        Footprints,
        ObsColumn,
        ObsSurface,
    )

    data_type_classes = {
        "surface": ObsSurface,
        "column": ObsColumn,
        "flux": Flux,
        # "met": ???
        "footprints": Footprints,
        "boundary_conditions": BoundaryConditions,
        "eulerian_model": EulerianModel,
    }

    return data_type_classes


def null_metadata_values() -> List:
    """
    Defines values which indicate metadata value is not specified.
    Returns:
        list: values to be seen as null
    """
    # TODO: Depending on how this is implemented, may want to update this to include np.nan values
    null_values = ["not_set", "NOT_SET"]

    return null_values



import logging
from pathlib import Path
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Union
import warnings
from xarray import Dataset

from openghg.util import species_lifetime, timestamp_now, check_function_open_nc
from openghg.store import infer_date_range, update_zero_dim
from openghg.types import multiPathType

logger = logging.getLogger("openghg.standardise.footprint")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def parse_acrg_org(
    filepath: multiPathType,
    site: str,
    domain: str,
    model: str,
    inlet: str,
    species: str,
    met_model: Optional[str] = None,
    network: Optional[str] = None,
    period: Optional[Union[str, tuple]] = None,
    continuous: bool = True,
    high_spatial_resolution: bool = False,
    time_resolved: bool = False,
    high_time_resolution: bool = False,
    short_lifetime: bool = False,
    chunks: Optional[Dict] = None,
) -> Dict:
    """
    Read and parse input emissions data in original ACRG format.

    Args:
        filepath: Path of file to load
        site: Site name
        domain: Domain of footprints
        model: Model used to create footprint (e.g. NAME or FLEXPART)
        inlet: Height above ground level in metres. Format 'NUMUNIT' e.g. "10m"
        met_model: Underlying meteorlogical model used (e.g. UKV)
        species: Species name. For a long-lived species this should be "inert".
        network: Network name
        period: Period of measurements. Only needed if this can not be inferred from the time coords
        continuous: Whether time stamps have to be continuous.
        high_spatial_resolution : Indicate footprints include both a low and high spatial resolution.
        time_resolved: Indicate footprints are high time resolution (include H_back dimension)
            Note this will be set to True automatically if species="co2" (Carbon Dioxide).
        high_time_resolution:  This argument is deprecated and will be replaced in future versions with time_resolved.
        short_lifetime: Indicate footprint is for a short-lived species. Needs species input.
            Note this will be set to True if species has an associated lifetime.
        chunks: Chunk schema to use when storing data the NetCDF. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
    Returns:
        dict: Dictionary of data
    """

    if high_time_resolution:
        warnings.warn(
            "This argument is deprecated and will be replaced in future versions with time_resolved.",
            DeprecationWarning,
        )
        time_resolved = high_time_resolution

    xr_open_fn, filepath = check_function_open_nc(filepath)

    with xr_open_fn(filepath).chunk(chunks) as fp_data:
        if chunks:
            logger.info(f"Rechunking with chunks={chunks}")

    if species == "co2":
        # Expect co2 data to have high time resolution
        if not time_resolved:
            logger.info("Updating time_resolved to True for co2 data")
            time_resolved = True

    if short_lifetime:
        if species == "inert":
            raise ValueError(
                "When indicating footprint is for short lived species, 'species' input must be included"
            )
    else:
        if species == "inert":
            lifetime = None
        else:
            lifetime = species_lifetime(species)
            if lifetime is not None:
                # TODO: May want to add a check on length of lifetime here
                logger.info("Updating short_lifetime to True since species has an associated lifetime")
                short_lifetime = True

    dv_rename = {
        # "fp": "srr",
        "temperature": "air_temperature",
        "pressure": "air_pressure",
        "wind_direction": "wind_from_direction",
        "PBLH": "atmosphere_boundary_layer_thickness",
    }

    attribute_rename = {"fp_output_units": "lpdm_native_output_units"}

    # # Removed for now - this renaming to match to PARIS would mean the dimension names
    # # were inconsistent between data types/
    # dim_rename = {"lat": "latitude", "lon": "longitude"}

    dim_drop = "lev"

    dim_reorder = ("time", "height", "lat", "lon")

    dv_attribute_updates: Dict[str, Dict[str, str]] = {}
    variable_names = [
        # "srr",
        "fp",
        "air_temperature",
        "air_pressure",
        "wind_speed",
        "wind_from_direction",
        "atmosphere_boundary_layer_thickness",
        "release_lon",
        "release_lat",
    ]

    for dv in variable_names:
        dv_attribute_updates[dv] = {}

    # dv_attribute_updates["srr"]["long_name"] = "source_receptor_relationship"
    dv_attribute_updates["fp"]["long_name"] = "source_receptor_relationship"
    dv_attribute_updates["air_temperature"]["long_name"] = "air temperature at release"
    dv_attribute_updates["air_pressure"]["long_name"] = "air pressure at release"
    dv_attribute_updates["atmosphere_boundary_layer_thickness"][
        "long_name"
    ] = "atmospheric boundary layer thickness at release"

    dv_attribute_updates["wind_speed"]["units"] = "m s-1"
    dv_attribute_updates["wind_speed"]["long_name"] = "wind speed at release"

    dv_attribute_updates["wind_from_direction"]["units"] = "degree"
    dv_attribute_updates["wind_from_direction"]["long_name"] = "wind direction at release"

    dv_attribute_updates["release_lon"]["units"] = "degree_east"
    dv_attribute_updates["release_lon"]["long_name"] = "Release longitude"
    dv_attribute_updates["release_lat"]["units"] = "degree_north"
    dv_attribute_updates["release_lat"]["long_name"] = "Release latitude"

    # Ignore type - dv_rename type should be fine as a dict but mypy unhappy.
    fp_data = fp_data.rename(**dv_rename)  # type: ignore
    # fp_data = fp_data.rename(**dim_rename)  # removed for now - see above

    fp_data = fp_data.drop_dims(dim_drop)
    fp_data = fp_data.transpose(*dim_reorder, ...)

    for attr, new_attr in attribute_rename.items():
        if attr in fp_data:
            fp_data.attrs[new_attr] = fp_data.attrs.pop(attr)

    for dv, attr_details in dv_attribute_updates.items():
        for key, value in attr_details.items():
            fp_data[dv].attrs[key] = value

    # Need to read the metadata from the footprints and then store it
    # Do we need to chunk the footprints / will a Datasource store it correctly?
    metadata: Dict[str, Union[str, float, List[float]]] = {}

    metadata["data_type"] = "footprints"
    metadata["site"] = site
    metadata["domain"] = domain
    metadata["model"] = model

    # Include both inlet and height keywords for backwards compatability
    metadata["inlet"] = inlet
    metadata["height"] = inlet
    metadata["species"] = species

    if met_model is not None:
        metadata["met_model"] = met_model

    if network is not None:
        metadata["network"] = network

    # Check if time has 0-dimensions and, if so, expand this so time is 1D
    if "time" in fp_data.coords:
        fp_data = update_zero_dim(fp_data, dim="time")

    fp_time = fp_data["time"]

    # If filepath is a single file, the naming scheme of this file can be used
    # as one factor to try and determine the period.
    # If multiple files, this input isn't needed.
    if isinstance(filepath, (str, Path)):
        input_filepath = filepath
    else:
        input_filepath = None

    start_date, end_date, period_str = infer_date_range(
        fp_time, filepath=input_filepath, period=period, continuous=continuous
    )

    metadata["start_date"] = str(start_date)
    metadata["end_date"] = str(end_date)
    metadata["time_period"] = period_str

    metadata["max_longitude"] = round(float(fp_data["lon"].max()), 5)
    metadata["min_longitude"] = round(float(fp_data["lon"].min()), 5)
    metadata["max_latitude"] = round(float(fp_data["lat"].max()), 5)
    metadata["min_latitude"] = round(float(fp_data["lat"].min()), 5)

    if high_spatial_resolution:
        try:
            metadata["max_longitude_high"] = round(float(fp_data["lon_high"].max()), 5)
            metadata["min_longitude_high"] = round(float(fp_data["lon_high"].min()), 5)
            metadata["max_latitude_high"] = round(float(fp_data["lat_high"].max()), 5)
            metadata["min_latitude_high"] = round(float(fp_data["lat_high"].min()), 5)

        except KeyError:
            raise KeyError("Expected high spatial resolution. Unable to find lat_high or lon_high data.")

    metadata["time_resolved"] = str(time_resolved)
    metadata["high_spatial_resolution"] = str(high_spatial_resolution)
    metadata["short_lifetime"] = str(short_lifetime)

    metadata["heights"] = [float(h) for h in fp_data.height.values]
    # Do we also need to save all the variables we have available in this footprints?
    metadata["variables"] = list(fp_data.data_vars)

    # if model_params is not None:
    #     metadata["model_parameters"] = model_params

    # TODO: Decide if to remove this as may not be the right thing for data we're outputting.
    # Set the attributes of this Dataset
    fp_data.attrs = {"author": "OpenGHG Cloud", "processed": str(timestamp_now())}

    # This might seem longwinded now but will help when we want to read
    # more than one footprints at a time
    # TODO - remove this once assign_attributes has been refactored
    key = "_".join((site, domain, model, inlet))

    footprint_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
    footprint_data[key]["data"] = fp_data
    footprint_data[key]["metadata"] = metadata

    return footprint_data



from collections import defaultdict
import copy
import logging
from typing import DefaultDict, Dict, List, Set, Optional, Union

from openghg.store.base import Datasource
from openghg.objectstore.metastore import open_metastore
from openghg.objectstore import get_writable_bucket, get_writable_buckets
from openghg.types import ObjectStoreError

logger = logging.getLogger("openghg.dataobjects")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class DataManager:
    def __init__(self, metadata: Dict[str, Dict], store: str):
        # We don't want the object store in this metadata as we want it to be the
        # unadulterated metadata to properly reflect what's stored.
        self.metadata = self._clean_metadata(metadata=metadata)
        self._store = store
        self._bucket = get_writable_bucket(name=store)
        self._backup: DefaultDict[str, Dict[str, Dict]] = defaultdict(dict)
        self._latest = "latest"

    def __str__(self) -> str:
        return str(self.metadata)

    def __bool__(self) -> bool:
        return bool(self.metadata)

    def _clean_metadata(self, metadata: Dict) -> Dict:
        """Ensures the metadata we give to the user is the metadata
        stored in the metastore and the Datasource and hasn't been modified by the
        search function. Currently this just removes the object_store key

        Args:
            metadata: Dictionary of metadata, we expect
        Returns:
            dict: Metadata without specific keys
        """
        metadata = copy.deepcopy(metadata)
        for m in metadata.values():
            try:
                del m["object_store"]
            except KeyError:
                pass

        return metadata

    def _check_datatypes(self, uuid: Union[str, List]) -> str:
        """Check the UUIDs are correct and ensure they all
        belong to a single data type

        Args:
            uuid: UUID(s) to check
        Returns:
            None
        """
        if not isinstance(uuid, list):
            uuid = [uuid]

        invalid_keys = [k for k in uuid if k not in self.metadata]

        if invalid_keys:
            raise ValueError(f"Invalid UUIDs: {invalid_keys}")

        # We should only have one data type
        data_types: Set[str] = {self.metadata[i]["data_type"] for i in uuid}

        if not data_types:
            raise ValueError("Unable to read data_type from metadata.")

        if len(data_types) > 1:
            raise ValueError(
                f"We can only modify Datasources of a single data type at once. We currently have {data_types}"
            )

        return data_types.pop()

    def refresh(self) -> None:
        """Force refresh the internal metadata store with data from the object store.

        Returns:
            None
        """
        from openghg.retrieve import search

        uuids = list(self.metadata.keys())
        res = search(uuid=uuids)
        # We don't want the object store in this metadata as we want it to be the
        # unadulterated metadata to properly reflect what's stored.
        for m in res.metadata.values():
            try:
                del m["object_store"]
            except KeyError:
                pass

        self.metadata = self._clean_metadata(metadata=res.metadata)

    def restore(self, uuid: str, version: Union[str, int] = "latest") -> None:
        """Restore a backed-up version of a Datasource's metadata.

        Args:
            uuid: UUID of Datasource to retrieve
            version: Version of metadata to restore
        Returns:
            None
        """
        if version == "latest":
            version = self._latest

        version = str(version)

        dtype = self._check_datatypes(uuid=uuid)
        with open_metastore(data_type=dtype, bucket=self._bucket) as metastore:
            backup = self._backup[uuid][version]
            self.metadata[uuid] = backup

            metastore.delete({"uuid": uuid})
            metastore.insert(backup)

            d = Datasource(bucket=self._bucket, uuid=uuid)
            d._metadata = backup
            d.save()

    def view_backup(self, uuid: Optional[str] = None, version: Optional[str] = None) -> Dict:
        """View backed-up metadata for all Datasources
        or a single Datasource if a UUID is passed in.

        Args:
            uuid: UUID of Datasource
        Returns:
            dict: Dictionary of versioned metadata
        """
        if uuid is not None:
            if version is not None:
                version = str(version)
                return self._backup[uuid][version]

            return self._backup[uuid]
        else:
            return self._backup

    def update_metadata(
        self,
        uuid: Union[List, str],
        to_update: Optional[Dict] = None,
        to_delete: Union[str, List, None] = None,
    ) -> None:
        """Update the metadata associated with data. This takes UUIDs of Datasources and updates
        the associated metadata. To update metadata pass in a dictionary of key/value pairs to update.
        To delete metadata pass in a list of keys to delete.

        Args:
            uuid: UUID(s) of Datasources to be updated.
            to_update: Dictionary of metadata to add/update. New key/value pairs will be added.
            If the key already exists in the metadata the value will be updated.
            to_delete: Key(s) to delete from the metadata
        Returns:
            None
        """
        if to_update is None and to_delete is None:
            return None

        if not isinstance(uuid, list):
            uuid = [uuid]

        dtype = self._check_datatypes(uuid=uuid)

        with open_metastore(bucket=self._bucket, data_type=dtype) as metastore:
            for u in uuid:
                updated = False
                d = Datasource(bucket=self._bucket, uuid=u)
                # Save a backup of the metadata for now
                found_record = metastore.search({"uuid": u})
                current_metadata = found_record[0]

                version = str(len(self._backup[u].keys()) + 1)
                self._latest = version
                self._backup[u][version] = copy.deepcopy(dict(current_metadata))
                # To update this object's records
                internal_copy = copy.deepcopy(dict(current_metadata))
                n_records = len(self._backup[u][version])

                # Do a quick check to make sure we're not being asked to delete all the metadata
                if to_delete is not None and to_delete:
                    if not isinstance(to_delete, list):
                        to_delete = [to_delete]

                    if "uuid" in to_delete:
                        raise ValueError("Cannot delete the UUID key.")

                    if len(to_delete) == n_records:
                        raise ValueError("We can't remove all the metadata associated with this Datasource.")
                    for k in to_delete:
                        d._metadata.pop(k)
                        internal_copy.pop(k)

                    try:
                        metastore.update(where={"uuid": u}, to_delete=to_delete)
                    except KeyError:
                        raise ValueError(
                            "Unable to remove keys from metadata store, please ensure they exist."
                        )

                    updated = True

                if to_update is not None and to_update:
                    if "uuid" in to_update:
                        raise ValueError("Cannot update the UUID.")

                    d._metadata.update(to_update)
                    internal_copy.update(to_update)
                    metastore.update(where={"uuid": u}, to_update=to_update)

                    updated = True

                if updated:
                    d.save()
                    # Update the metadata stored internally so we're up to date
                    self.metadata[u] = internal_copy
                    logger.info(f"Modified metadata for {u}.")

    def delete_datasource(self, uuid: Union[List, str]) -> None:
        """Delete Datasource(s) in the object store.
        At the moment we only support deleting the complete Datasource.

        NOTE: Make sure you really want to delete the Datasource(s)

        Args:
            uuid: UUID(s) of objects to delete
        Returns:
            None
        """
        from openghg.objectstore import delete_object

        # Add in ability to delete metadata keys
        if not isinstance(uuid, list):
            uuid = [uuid]

        dtype = self._check_datatypes(uuid=uuid)

        with open_metastore(bucket=self._bucket, data_type=dtype) as metastore:
            for uid in uuid:
                # First remove the data from the metadata store
                metastore.delete({"uuid": uid})

                # Delete all the data associated with a Datasource and the
                # data in its zarr store.
                d = Datasource(bucket=self._bucket, uuid=uid)
                d.delete_all_data()

                # Then delete the Datasource itself
                delete_object(bucket=self._bucket, key=d.key())

                logger.info(f"Deleted Datasource with UUID {uid}.")


def data_manager(data_type: str, store: str, **kwargs: Dict) -> DataManager:
    """Lookup the data / metadata you'd like to modify.

    Args:
        data_type: Type of data, for example surface, flux, footprint
        store: Name of store
        kwargs: Any pair of keyword arguments for searching
    Returns:
        DataManager: A handler object to help modify the metadata
    """
    from openghg.dataobjects import DataManager
    from openghg.retrieve import search

    writable_stores = get_writable_buckets()

    if store not in writable_stores:
        raise ObjectStoreError(f"You do not have permission to write to the {store} store.")

    res = search(data_type=data_type, store=store, **kwargs)
    metadata = res.metadata
    return DataManager(metadata=metadata, store=store)



"""Utility functions for printing version information.

This code was slightly modified from the xarray version

See https://github.com/pydata/xarray/blob/main/xarray/util/print_versions.py

Thank you xarray devs.
"""

import importlib
import locale
import os
import platform
import struct
import subprocess
import sys
from typing import List, IO, Union, Tuple

__all__ = ["show_versions", "check_if_need_new_version"]


def get_sys_info() -> List:
    """Returns system information as a list

    Returns:
        list: List of system information
    """
    blob: List[Tuple] = []

    # get full commit hash
    commit: Union[str, bytes, None] = None
    if os.path.isdir(".git") and os.path.isdir("openghg"):
        try:
            pipe = subprocess.Popen(
                'git log --format="%H" -n 1'.split(" "),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            so, _ = pipe.communicate()
        except Exception:
            pass
        else:
            if pipe.returncode == 0:
                commit = so
                try:
                    commit = so.decode("utf-8")
                except ValueError:
                    pass

                commit = commit.strip().strip('"')  # type: ignore

    blob.append(("commit", commit))

    try:
        (sysname, _, release, _, machine, processor) = platform.uname()
        blob.extend(
            [
                ("python", sys.version),
                ("python-bits", struct.calcsize("P") * 8),
                ("OS", f"{sysname}"),
                ("OS-release", f"{release}"),
                # ("Version", f"{version}"),
                ("machine", f"{machine}"),
                ("processor", f"{processor}"),
                ("byteorder", f"{sys.byteorder}"),
                ("LC_ALL", f'{os.environ.get("LC_ALL", "None")}'),
                ("LANG", f'{os.environ.get("LANG", "None")}'),
                ("LOCALE", f"{locale.getlocale()}"),
            ]
        )
    except Exception:
        pass

    return blob


def netcdf_and_hdf5_versions() -> List:
    """Returns the versions of NetCDF and HDF5 libraries installed.

    Returns:
        list: List of versions
    """
    libhdf5_version = None
    libnetcdf_version = None
    try:
        import netCDF4

        libhdf5_version = netCDF4.__hdf5libversion__
        libnetcdf_version = netCDF4.__netcdf4libversion__
    except ImportError:
        pass

    return [("libhdf5", libhdf5_version), ("libnetcdf", libnetcdf_version)]


def show_versions(file: IO = sys.stdout) -> None:
    """print the versions of xarray and its dependencies

    Args:
        file : file-like, optional
            print to the given file-like object. Defaults to sys.stdout.
    """
    sys_info = get_sys_info()

    try:
        sys_info.extend(netcdf_and_hdf5_versions())
    except Exception as e:
        print(f"Error collecting netcdf / hdf5 version: {e}")

    deps = [
        # (MODULE_NAME, f(mod) -> mod version)
        ("addict", lambda mod: mod.__version__),
        ("dask", lambda mod: mod.__version__),
        ("h5netcdf", lambda mod: mod.__version__),
        ("icoscp", lambda mod: mod.__version__),
        ("matplotlib", lambda mod: mod.__version__),
        ("msgpack", lambda mod: mod.__version__),
        ("netcdf4", lambda mod: mod.__version__),
        ("nbformat", lambda mod: mod.__version__),
        ("numexpr", lambda mod: mod.__version__),
        ("numpy", lambda mod: mod.__version__),
        ("nc-time-axis", lambda mod: mod.__version__),
        ("pandas", lambda mod: mod.__version__),
        ("plotly", lambda mod: mod.__version__),
        ("pyvis", lambda mod: mod.__version__),
        ("rapidfuzz", lambda mod: mod.__version__),
        ("requests", lambda mod: mod.__version__),
        ("scipy", lambda mod: mod.__version__),
        ("tinydb", lambda mod: mod.__version__),
        ("toml", lambda mod: mod.__version__),
        ("rich", lambda mod: mod.__version__),
        ("xarray", lambda mod: mod.__version__),
        ("urllib3", lambda mod: mod.__version__),
        # openghg setup/test/doc
        ("setuptools", lambda mod: mod.__version__),
        ("pip", lambda mod: mod.__version__),
        ("conda", lambda mod: mod.__version__),
        ("pytest", lambda mod: mod.__version__),
        ("IPython", lambda mod: mod.__version__),
        ("sphinx", lambda mod: mod.__version__),
    ]

    deps_blob: List[Tuple] = []
    for modname, ver_f in deps:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = importlib.import_module(modname)
        except Exception:
            deps_blob.append((modname, None))
        else:
            try:
                ver = ver_f(mod)
                deps_blob.append((modname, ver))
            except Exception:
                deps_blob.append((modname, "installed"))

    print("\nINSTALLED VERSIONS", file=file)
    print("------------------", file=file)

    for k, stat in sys_info:
        print(f"{k}: {stat}", file=file)

    print("", file=file)
    for k, stat in deps_blob:
        print(f"{k}: {stat}", file=file)


def check_if_need_new_version(if_exists: str = "auto", save_current: str = "auto") -> bool:
    """
    Check combination of if_exists and save_current keywords to determine
    whether a new version should be created.

    Output related to these parameters:
        - if_exists="auto", save_current="auto"
           - new_version=False (default) - If both values are set
             to "auto", data will only be updated if there is no
             overlapping data. In this case we can safely write
             to the same version with no data conflict.
        - if_exists="combine"/"new", save_current="auto"
           - new_version=True - If a scheme has been set for the
             combination of new and current data and save_current is "auto",
             create a new version.
        - if_exists="auto"/"combine"/"new", save_current="y"/"yes"
           - new_version=True - If save_current is explicitly set
             to "y", create a new version.
        - if_exists="auto"/"combine"/"new", save_current="n"/"no"
           - new_version=False - If save_current is explicitly set
             to "n", allow previous version to be overwritten.

    Args:
        if_exists: How to combine new and current data, if present.
        save_current: Whether to save current data or replace this.
    Returns:
        bool: Whether new version should be created
    """
    # Determining whether a new version should be created based on inputs.
    if if_exists == "auto" and save_current == "auto":
        # Add new (non-overlapping) data on the same version
        new_version = False
    elif if_exists != "auto" and save_current == "auto":
        # If data could be modified based on if_exists input
        # default to creating a new version.
        new_version = True
    elif save_current in ["y", "yes"]:
        # Otherwise match new version to the save_current input.
        new_version = True
    elif save_current in ["n", "no"]:
        # Otherwise match new version to the save_current input.
        new_version = False

    return new_version



from ._bc_data import BoundaryConditionsData
from ._flux_data import FluxData
from ._footprint_data import FootprintData
from ._metdata import METData
from ._obsdata import ObsData
from ._obscolumn_data import ObsColumnData
from ._searchresults import SearchResults
from ._datamanager import DataManager, data_manager



import argparse


def cli() -> None:
    from openghg.util import create_config, show_versions

    parser = argparse.ArgumentParser(
        prog="OpenGHG CLI",
        description="The OpenGHG Command Line Interface helps you get OpenGHG setup on your local machine.",
        epilog="Text at the bottom of help",
    )
    parser.add_argument(
        "--default-config", action="store_true", help="Get OpenGHG setup with default a configuration"
    )
    parser.add_argument("--quickstart", action="store_true", help="Run the quickstart setup process")
    parser.add_argument("--version", action="store_true", help="Print the version information about OpenGHG")

    args = parser.parse_args()

    if args.default_config:
        create_config(silent=True)
    elif args.quickstart:
        create_config()
    elif args.version:
        show_versions()
    else:
        parser.print_help()



import logging
import os
import platform
from pathlib import Path
from typing import Dict, Optional
import uuid
import toml
import shutil
from openghg.types import ConfigFileError

logger = logging.getLogger("openghg.util")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler

openghg_config_filename = "openghg.conf"


# @lru_cache
def get_user_id() -> str:
    """Return the user's ID

    Returns:
        str: User ID
    """
    config = read_local_config()
    uid = str(config.get("user_id", "NA"))
    return uid


def get_default_objectstore_path() -> Path:
    """Returns the default object store path in the user's home directory

    Returns:
        Path: Object store path in ~/openghg_store
    """
    return Path.home().joinpath("openghg_store").absolute()


# @lru_cache
def get_user_config_path() -> Path:
    """Returns path to user config file.

    This file is created in the user's home directory
    in  ~/.ghgconfig/openghg/user.conf on Linux / macOS or
    in LOCALAPPDATA/openghg/openghg.conf on Windows.

    Returns:
        pathlib.Path: Path to user config file
    """
    user_platform = platform.system()

    if user_platform == "Windows":
        appdata_path = os.getenv("LOCALAPPDATA")
        if appdata_path is None:
            raise ValueError("Unable to read LOCALAPPDATA environment variable.")

        config_path = Path(appdata_path).joinpath("openghg", openghg_config_filename)
    elif user_platform in ("Linux", "Darwin"):
        config_path = Path.home().joinpath(".openghg", openghg_config_filename)
    else:
        raise ValueError(f"Unknown platform: {user_platform}")

    return config_path


def create_config(silent: bool = False) -> None:
    """Creates a user config.

    Args:
        silent: Creates the basic configuration file with only
        the user's object store in a default location.
    Returns:
        None
    """
    if not silent:
        print("\nOpenGHG configuration")
        print("---------------------\n")

    user_config_path = get_user_config_path()

    # Current config version as of version 0.6.0
    config_version = "2"
    positive_responses = ("y", "yes")

    # If the config file exists we might need to update it due to the introduction
    # of the user ID and new object store path handling for multiple stores
    if user_config_path.exists():
        if silent:
            logger.error("Cannot update an existing configuration silently. Please run interactively.")
            return

        logger.info(f"User config exists at {str(user_config_path)}, checking...")

        config = toml.loads(user_config_path.read_text())

        recent = "config_version" in config

        if recent:
            user_store_path = Path(config["object_store"]["user"]["path"])
        else:
            user_store_path = Path(config["object_store"]["local_store"])

        logger.info(f"Current user object store path: {user_store_path}")

        # Store the object store info
        stores = {}

        update_input = input("Would you like to update the path? (y/n): ")
        if update_input.lower() in positive_responses:
            new_path_input = input("Enter new path for object store: ")

            if not new_path_input:
                print("You must enter a path. Unable to complete config setup.")
                return

            new_path = Path(new_path_input).expanduser().resolve()
            stores["user"] = {"path": str(new_path), "permissions": "rw"}
        else:
            stores["user"] = {"path": str(user_store_path), "permissions": "rw"}

        # Copy in exisiting shared stores
        if recent:
            stores.update({k: v for k, v in config["object_store"].items() if k != "user"})

        # Now ask the user if they want to add new stores
        new_shared_stores = _user_multstore_input()

        if new_shared_stores:
            existing = [k for k in new_shared_stores if k in stores]

            existing_paths = [path_data["path"] for path_data in stores.values()]

            # Here it checks if newly entered paths are already present in config file.
            duplicate_paths_with_store_name = {
                store_name: path_data
                for store_name, path_data in new_shared_stores.items()
                if path_data["path"] in existing_paths
            }

            if existing:
                print(f"Some names match those of existing stores: {existing}, please update manually.")

            if duplicate_paths_with_store_name:
                raise ValueError(
                    f"Paths of the following new stores match those of existing store: {duplicate_paths_with_store_name}"
                )

            stores.update(new_shared_stores)

        # Some users may not have a user ID if they've used previous versions of OpenGHG
        user_id = config.get("user_id")
        config = _combine_config(config_version=config_version, object_stores=stores, user_id=user_id)
    else:
        # Let's try migrating the old config
        # If it works we call this function again
        # otherwise continue on to create a new config
        try:
            _migrate_config()
        except FileNotFoundError:
            pass
        else:
            create_config(silent=silent)
            return

        stores = {}

        # 1. Create the user's object store first
        if not silent:
            logger.info("We'll first create your user object store.\n")

        obj_store_path = get_default_objectstore_path()

        if silent:
            obj_store_path = get_default_objectstore_path()
        else:
            obj_store_input = input(f"Enter path for your local object store (default {obj_store_path}): ")
            if obj_store_input:
                obj_store_path = Path(obj_store_input).expanduser().resolve()

        # Let's create the store to make sure it's a valid path
        obj_store_path.mkdir(parents=True, exist_ok=True)

        stores["user"] = {"path": str(obj_store_path), "permissions": "rw"}

        if not silent:
            shared_stores = _user_multstore_input()
            stores.update(shared_stores)

        config = _combine_config(config_version=config_version, object_stores=stores)

    # Make the .config/openghg folder
    user_config_path.parent.mkdir(parents=True, exist_ok=True)

    if not silent:
        logger.info(f"Configuration written to {user_config_path}")

    user_config_path.write_text(toml.dumps(config))


def _user_multstore_input() -> Dict:
    """Ask the user to input data about shared object stores

    Returns:
        dict: Dictionary of object store paths and permissions
    """
    positive_responses = ("y", "yes")
    stores = {}
    # 2. Ask the user to enter other object store paths
    while True:
        response = input("Would you like to add another object store? (y/n): ")
        if response.lower() in positive_responses:
            store_name = input("Enter the name of the store: ")
            store_path = input("Enter the object store path: ")
            print("\nYou will now be asked for read/write permissions for the store.")
            print("For read only enter r, for read and write enter rw.")

            store_permissions = ""
            while store_permissions not in ("r", "rw"):
                store_permissions = input("\nEnter object store permissions: ")

            stores[store_name] = {"path": store_path, "permissions": store_permissions}
        else:
            break

    return stores


def _combine_config(config_version: str, object_stores: Dict, user_id: Optional[str] = None) -> Dict:
    """Combine parts required into the proper dictionary format

    Args:
        config_version: Configuration version number
        object_stores: Object store configuration dictionary
        user_id: User ID
    """
    # Create the object store dictionary
    object_store_info = {}
    for name, data in object_stores.items():
        path = str(Path(data["path"]).expanduser().resolve())
        permissions = data["permissions"].strip()

        object_store_info[name] = {"path": path, "permissions": permissions}

    if user_id is None:
        user_id = str(uuid.uuid4())

    return {"user_id": user_id, "config_version": config_version, "object_store": object_store_info}


# @lru_cache
def read_local_config() -> Dict:
    """Reads the local config file.

    Returns:
        dict: OpenGHG configurations
    """
    config_path = get_user_config_path()

    if not config_path.exists():
        try:
            _migrate_config()
        except FileNotFoundError as e:
            raise ConfigFileError(
                "Unable to read configuration file, please see the installation instructions \
                or run openghg --quickstart"
            ) from e

    config: Dict = toml.loads(config_path.read_text())

    try:
        _ = config["object_store"]["user"]
    except KeyError:
        raise ConfigFileError(
            "Invalid config file detected, please please see the installation instructions \
                or run openghg --quickstart"
        )

    # Check see is the store uses the new zarr storage format
    # for OpenGHG >= 0.8.0
    valid_stores = {}
    for name, store_data in config["object_store"].items():
        store_path = Path(store_data["path"])
        # If it doesn't exist or its empty then we expect it to be created / populated
        if not store_path.exists() or not any(store_path.iterdir()):
            valid_stores[name] = store_data
        # Otherwise we check an existing store to see if it's the correct format
        else:
            if _check_valid_store(store_path):
                valid_stores[name] = store_data
            else:
                logger.warning(
                    f"Object store {name} does not use the new Zarr storage format and will be ignored."
                )

    if not valid_stores:
        raise ConfigFileError(
            "We've only detected old (non-Zarr) object stores. Please update your configuration to add a new path."
        )

    config["object_store"] = valid_stores

    return config


def check_config() -> None:
    """Check that the user config file is valid and the paths
    given in it exist. Raises ConfigFileError if problems found.

    Returns:
        None
    """
    config_path = get_user_config_path()
    please_update = "please run openghg --quickstart to update it."

    if not config_path.exists():
        raise ConfigFileError(
            "Configuration file does not exist. Please create it by running openghg --quickstart."
        )

    config = read_local_config()
    try:
        uid = config["user_id"]
    except KeyError:
        raise ConfigFileError("Unable to read user ID, ")

    try:
        uuid.UUID(uid, version=4)
    except ValueError:
        raise ConfigFileError(f"Invalid user ID, {please_update}")

    try:
        _ = config["config_version"]
    except KeyError:
        raise ConfigFileError(f"Invalid config file, {please_update}")

    try:
        object_stores = config["object_store"]
    except KeyError:
        raise ConfigFileError(f"Unable to read object store data, {please_update}")

    for name, data in object_stores.items():
        p = Path(data["path"])
        if not p.exists():
            logger.info(f"The path for object store {name} at {p} does not exist but will be created.")


def _migrate_config() -> None:
    """If user config file is in ~/.config, move it to ~/.openghg.

    If no config is found in ~/.config or system is Windows, raise FileNotFoundError.

    Returns:
        None
    """
    old_config_path = Path.home().joinpath(".config", "openghg", openghg_config_filename)

    if old_config_path.exists():
        new_config_path = get_user_config_path()
        new_config_path.parent.mkdir(parents=True)
        shutil.move(str(old_config_path), str(new_config_path))
        logger.info(f"Moved user config file from {str(old_config_path)} to {str(new_config_path)}.")
        shutil.rmtree(old_config_path.parent)  # remove "openghg" dir from ~/.config
    else:
        raise FileNotFoundError("Configuration file not found.")


def _check_valid_store(store_path: Path) -> bool:
    """Checks if the store is a valid object store using the new Zarr storage
    format. If it is return True, otherwise False.

    TODO - remove this when users have all moved to the new object store format.

    Args:
        store_path: Object store path
    Returns:
        bool: True if valid, False if not
    """
    data_dir = Path(store_path).joinpath("data")
    # Now check if there's a zarr folder in the data directory
    store_dirs = list(data_dir.glob("*"))
    # Let's take the first data directory and see if there's a zarr folder in it
    if not store_dirs:
        logger.info(
            f"No data found in the object store {store_path}, "
            "so we are treating this empty store as a zarr store."
        )
        return True

    store_data_dir = store_dirs[0]

    return store_data_dir.joinpath("zarr").exists()



import bz2
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Optional, Union

from openghg.types import pathType, multiPathType

__all__ = [
    "load_parser",
    "load_surface_parser",
    "load_column_parser",
    "load_column_source_parser",
    "load_flux_parser",
    "load_flux_database_parser",
    "get_datapath",
    "get_logfile_path",
    "load_json",
    "read_header",
    "compress",
    "decompress",
    "compress_str",
    "decompress_str",
    "compress_json",
    "decompress_json",
]


def load_parser(data_name: str, module_name: str) -> Callable:
    """Load parse function from within module.

    This expects a function of the form to be:
        - parse_{data_name}()
    and for this to have been imported with an appropriate __init__.py module.

    Args:
        data_name: Name of data type / database / data source for the
        parse function.
        module_name: Full module name to be imported e.g. "openghg.standardise.surface"
    Returns:
        Callable : parse function
    """
    from importlib import import_module

    module = import_module(name=module_name)

    function_name = f"parse_{data_name.lower()}"
    fn: Callable = getattr(module, function_name)

    return fn


def load_surface_parser(source_format: str) -> Callable:
    """Load parsing object for the obssurface data type.
    Used with `openghg.standardise.surface` sub-module

    Args:
        source_format: Name of data type such as CRDS
    Returns:
        callable: class_name object
    """
    surface_module_name = "openghg.standardise.surface"
    fn = load_parser(data_name=source_format, module_name=surface_module_name)

    return fn


def load_column_parser(source_format: str) -> Callable:
    """Load a parsing object for the obscolumn data type.
    Used with `openghg.standardise.column` sub-module

    Args:
        source_format: Name of data type e.g. OPENGHG
    Returns:
        callable: parser function
    """
    column_st_module = "openghg.standardise.column"
    fn = load_parser(data_name=source_format, module_name=column_st_module)

    return fn


def load_column_source_parser(source_format: str) -> Callable:
    """Load a parsing object for the source of column data.
    Used with `openghg.transform.column` sub-module

    Args:
        source_format: Name of data source e.g. GOSAT
    Returns:
        callable: parser function
    """
    column_tr_module = "openghg.transform.column"
    fn = load_parser(data_name=source_format, module_name=column_tr_module)

    return fn


def load_flux_parser(source_format: str) -> Callable:
    """Load a parsing object for the "flux" data type.
    Used with `openghg.standardise.flux` sub-module

    Args:
        source_format: Name of data type e.g. OPENGHG
    Returns:
        callable: parser function
    """
    flux_st_module_name = "openghg.standardise.flux"
    fn = load_parser(data_name=source_format, module_name=flux_st_module_name)

    return fn


def load_flux_database_parser(database: str) -> Callable:
    """Load a parsing object for the source of column data.
    Used with `openghg.transform.flux` sub-module

    Args:
        database: Name of data source e.g. EDGAR
    Returns:
        callable: parser function
    """
    flux_tr_module_name = "openghg.transform.flux"
    fn = load_parser(data_name=database, module_name=flux_tr_module_name)

    return fn


def load_footprint_parser(source_format: str) -> Callable:
    """Load a parsing object for the footprint data type.
    Used with `openghg.standardise.footprint` sub-module

    Args:
        source_format: Name of data type e.g. OPENGHG, ACRG_ORG
    Returns:
        callable: parser function
    """
    footprint_st_module_name = "openghg.standardise.footprint"
    fn = load_parser(data_name=source_format, module_name=footprint_st_module_name)

    return fn


def get_datapath(filename: pathType, directory: Optional[str] = None) -> Path:
    """Returns the correct path to data files used for assigning attributes

    Args:
        filename: Name of file to be accessed
    Returns:
        pathlib.Path: Path of file
    """
    from pathlib import Path

    filename = str(filename)

    if directory is None:
        return Path(__file__).resolve().parent.parent.joinpath(f"data/{filename}")
    else:
        return Path(__file__).resolve().parent.parent.joinpath(f"data/{directory}/{filename}")


def load_json(path: Union[str, Path]) -> Dict:
    """Returns a dictionary deserialised from JSON.

    Args:
        path: Path to file, can be any filepath
    Returns:
        dict: Dictionary created from JSON
    """
    with open(path, "r") as f:
        data: Dict[str, Any] = json.load(f)

    return data


def load_internal_json(filename: str) -> Dict:
    """Returns a dictionary deserialised from JSON. Pass filename to load data from JSON files in the
    openghg/data directory or pass a full filepath to path to load from any file.

    Args:
        filename: Name of JSON file. Must be located in openghg/data
        path: Path to file, can be any filepath
    Returns:
        dict: Dictionary created from JSON
    """
    file_path = get_datapath(filename=filename)
    return load_json(path=file_path)


def read_header(filepath: pathType, comment_char: str = "#") -> List:
    """Reads the header lines denoted by the comment_char

    Args:
        filepath: Path to file
        comment_char: Character that denotes a comment line
        at the start of a file
    Returns:
        list: List of lines in the header
    """
    comment_char = str(comment_char)

    header = []
    # Get the number of header lines
    with open(filepath, "r") as f:
        for line in f:
            if line.startswith(comment_char):
                header.append(line)
            else:
                break

    return header


def compress(data: bytes) -> bytes:
    """Compress the given data

    Args:
        data: Binary data
    Returns:
        bytes: Compressed data
    """
    return bz2.compress(data=data)


def decompress(data: bytes) -> bytes:
    """Decompress the given data

    Args:
        data: Compressed data
    Returns:
        bytes: Decompressed data
    """
    return bz2.decompress(data=data)


def compress_str(s: str) -> bytes:
    """Compress a string

    Args:
        s: String
    Retruns:
        bytes: Compressed data
    """
    return compress(data=s.encode(encoding="utf-8"))


def decompress_str(data: bytes) -> str:
    """Decompress a string from bytes

    Args:
        data: Compressed data
    Returns:
        str: Decompressed str
    """
    return decompress(data=data).decode(encoding="utf-8")


def decompress_json(data: bytes) -> Any:
    """Decompress a string and load to JSON

    Args:
        data: Compressed binary data
    Returns:
        Object loaded from JSON
    """
    decompressed = decompress_str(data=data)
    return json.loads(decompressed)


def compress_json(data: Any) -> bytes:
    """Convert object to JSON string and compress

    Args:
        data: Object to pass to json.dumps
    Returns:
        bytes: Compressed binary data
    """
    json_str = json.dumps(data)
    return compress_str(json_str)


def get_logfile_path() -> Path:
    """Get the logfile path

    Returns:
        Path: Path to logfile
    """
    from openghg.util import running_locally

    if running_locally():
        return Path.home().joinpath("openghg.log")
    else:
        return Path("/tmp/openghg.log")


def check_function_open_nc(filepath: multiPathType) -> Tuple[Callable, multiPathType]:
    """
    Check the filepath input to choose which xarray open function to use:
     - Path or single item List - use open_dataset
     - multiple item List - use open_mfdataset

    Args:
        filepath: Path or list of filepaths
    Returns:
        Callable, Union[Path, List[Path]]: function and suitable filepath
            to use with the function.
    """
    import xarray as xr

    if isinstance(filepath, list):
        if len(filepath) > 1:
            xr_open_fn: Callable = xr.open_mfdataset
        else:
            xr_open_fn = xr.open_dataset
            filepath = filepath[0]
    else:
        xr_open_fn = xr.open_dataset

    return xr_open_fn, filepath



from typing import Any, Dict, Tuple, List, Optional

import numpy as np
from numpy import ndarray

from openghg.types import optionalPathType, ArrayLikeMatch, ArrayLike, XrDataLike, XrDataLikeMatch

__all__ = ["get_domain_info", "find_domain", "convert_lon_to_180", "convert_lon_to_360"]


def get_domain_info(domain_filepath: optionalPathType = None) -> Dict[str, Any]:
    """Extract data from domain info JSON file as a dictionary.

    This uses the data stored within openghg_defs/domain_info JSON file by default.

    Args:
        domain_filepath: Alternative domain info file.
    Returns:
        dict: Data from domain JSON file
    """
    from openghg_defs import domain_info_file
    from openghg.util import load_json

    if domain_filepath is None:
        domain_info_json = load_json(path=domain_info_file)
    else:
        domain_info_json = load_json(path=domain_filepath)

    return domain_info_json


def find_domain(domain: str, domain_filepath: optionalPathType = None) -> Tuple[ndarray, ndarray]:
    """Finds the latitude and longitude values in degrees associated
    with a given domain name.

    Args:
        domain: Pre-defined domain name
        domain_filepath: Alternative domain info file. Defaults to openghg_defs input.
    Returns:
        array, array : Latitude and longitude values for the domain in degrees.
    """

    domain_info = get_domain_info(domain_filepath)

    # Look for domain in domain_info file
    if domain in domain_info:
        domain_data = domain_info[domain]
    elif domain.upper() in domain_info:
        domain = domain.upper()
        domain_data = domain_info[domain]
    else:
        raise ValueError(f"Pre-defined domain '{domain}' not found")

    # Extract or create latitude and longitude data
    latitude = _get_coord_data("latitude", domain_data, domain)
    longitude = _get_coord_data("longitude", domain_data, domain)

    return latitude, longitude


def _get_coord_data(coord: str, data: Dict[str, Any], domain: str) -> ndarray:
    """Attempts to extract or derive coordinate (typically latitude/longitude)
    values for a domain from provided data dictionary (typically
    this can be derived from 'domain_info.json' file).

    This looks for:
     - File containing coordinate values (in degrees)
         - Looks for "{coord}_file" attribute e.g. "latitude_file"
         - OR for a file within "domain" subfolder called "{domain}_{coord}.dat"
           e.g. "EUROPE_latitude.dat"
     - "{coord}_range" and "{coord}_increment" attributes to use to construct
       the coordinate values e.g. "latitude_range" to include the start and
       end (inclusive) range and "latitude_increment" for the step in degrees.

    Args:
        coord: Name of coordinate (e.g. latitude, longitude)
        data: Data dictionary containing details of domain
              (e.g. derived from 'domain_info.json')
        domain: Name of domain
    Returns:
        array: Extracted or derived coordinate values
    """
    from openghg_defs import data_path

    # Look for explicit file keyword in data e.g. "latitude_file"
    # Extract data from file if found and return
    filename_str = f"{coord}_file"
    if filename_str in data:
        full_filename = data_path / data[filename_str]
        coord_data: ndarray = np.loadtxt(full_filename)
        return coord_data

    # If no explicit file name defined, look within known location to see
    # if data is present by looking for file of form "domain/{domain}_{coord}.csv"
    # e.g. "domain/EUROPE_latitude.csv" (within "openghg/openghg/data" folder)
    try:
        full_filename = data_path / "domain" / f"{domain}_{coord}.dat"
        coord_data = np.loadtxt(full_filename)
    except OSError:
        pass
    else:
        return coord_data

    # If no data files can be found, look for coordinate range and increment values
    # If present, create the coordinate data. If not raise a ValueError.
    try:
        coord_range = data[f"{coord}_range"]
        increment = data[f"{coord}_increment"]
    except KeyError:
        raise ValueError(f"Unable to get {coord} coordinate data for domain: {domain}")

    coord_min = float(coord_range[0])
    coord_max = float(coord_range[-1])
    increment = float(increment)

    coord_data = np.arange(coord_min, coord_max + increment, increment)

    return coord_data


def find_coord_name(data: XrDataLike, options: List[str]) -> Optional[str]:
    """
    Find the name of a coordinate based on input options.
    Only the first found value will be returned.

    Args:
        data: xarray Data structure
        options: List of options to check. Will be checked in order.

    Returns:
        str / None: Name of coordinate if located within data. None otherwise.
    """
    for option in options:
        if option in data.coords:
            name = option
            break
    else:
        return None

    return name


def convert_lon_to_180(longitude: ArrayLikeMatch) -> ArrayLikeMatch:
    """
    Convert longitude extent from (0 to 360) to (-180 to 180).
    This does *not* reorder the values.

    Args:
        longitude: Valid longitude values in degrees.
    Returns:
        ndarray / DataArray : Updated longitude values in the same order.
    """

    # Check range of longitude values and convert to -180 to 180.
    longitude = ((longitude - 180) % 360) - 180

    return longitude


def convert_lon_to_360(longitude: ArrayLikeMatch) -> ArrayLikeMatch:
    """
    Convert longitude extent to (0 to 360).
    This does *not* reorder the values.

    Args:
        longitude: Valid longitude values in degrees.
    Returns:
        ndarray / DataArray : Updated longitude values in the same order.
    """

    # Check range of longitude values and convert to 0 to 360.
    longitude = longitude % 360

    return longitude


def convert_internal_longitude(
    data: XrDataLikeMatch, lon_name: Optional[str] = None, reorder: bool = True
) -> XrDataLikeMatch:
    """
    Convert longitude coordinate within an xarray data structure (DataArray or Dataset).

    Args:
        data: Data with longitude values to convert.
        lon_name: By default will look a coord called "lon" or "longitude".
            Otherwise must be specified.
        reorder: Whether to reorder the data based on the converted longitude values.
    Returns:
        DataArray / Dataset: Input data with updated longitude values
    """
    if lon_name is None:
        lon_options = ["lon", "longitude"]
        lon_name = find_coord_name(data, lon_options)
        if lon_name is None:
            raise ValueError("Please specify 'lon_name'.")

    longitude = data[lon_name]
    longitude = convert_lon_to_180(longitude)

    data = data.assign_coords({lon_name: longitude})

    if reorder:
        data = data.sortby(lon_name)

    return data


def cut_data_extent(
    data: XrDataLikeMatch,
    lat_out: ArrayLike,
    lon_out: ArrayLike,
    lat_name: Optional[str] = None,
    lon_name: Optional[str] = None,
    copy: bool = False,
) -> XrDataLikeMatch:
    """
    Cut down extent of data within an xarray data structure (DataArray or Dataset)
    against an output latitude and longitude range.

    A buffer based on the maximum difference along the lon_out and lat_out dimensions
    will be added when the data is cut.

    Args:
        data: Data to be cut down
        lat_out: Array containing output latitude values
        lon_out: Array containing output longitude values
        lat_name: Name of latitude dimension. Must be specified if not "lat" or "latitude".
        lon_name: Name of longitude dimension. Must be specified if not "lon" or "longitude".
        copy: Whether to explicitly copy the data.

    Returns:
        xarray.DataArray / xarray.Dataset: data with reduced lat, lon ranges.
    """
    if lat_name is None:
        lat_options = ["lat", "latitude"]
        lat_name = find_coord_name(data, lat_options)
        if lat_name is None:
            raise ValueError("Please specify 'lat_name'.")

    if lon_name is None:
        lon_options = ["lon", "longitude"]
        lon_name = find_coord_name(data, lon_options)
        if lon_name is None:
            raise ValueError("Please specify 'lon_name'.")

    if isinstance(lat_out, np.ndarray):
        lat_out.sort()
    else:
        lat_out = lat_out.sortby(lat_out[lat_name])

    if isinstance(lon_out, np.ndarray):
        lon_out.sort()
    else:
        lon_out = lon_out.sortby(lon_out[lon_name])

    lat_diff = (lat_out[1:] - lat_out[:-1]).max()
    lon_diff = (lon_out[1:] - lon_out[:-1]).max()

    lat_low = np.min(lat_out) - lat_diff
    lat_high = np.max(lat_out) + lat_diff
    lon_low = np.min(lon_out) - lon_diff
    lon_high = np.max(lon_out) + lon_diff

    lat_cut_wide_range = slice(lat_low, lat_high)
    lon_cut_wide_range = slice(lon_low, lon_high)

    if copy:
        data_cut = data.copy()
    else:
        data_cut = data

    data_cut = data_cut.sel({lat_name: lat_cut_wide_range, lon_name: lon_cut_wide_range})

    return data_cut



from ._chunking import ChunkingSchema
from ._store import Store
from ._localzarrstore import LocalZarrStore
from ._encoding import get_zarr_encoding
from ._compression import compare_compression
from ._convert import convert_store



from typing import Optional, Union, cast, overload
import logging
from openghg.types import optionalPathType

__all__ = ["format_inlet", "extract_height_name"]

logger = logging.getLogger("openghg.util")
logger.setLevel(logging.INFO)  # Have to set level for logger as well as handler


@overload
def format_inlet(
    inlet: str,
    units: str = "m",
    key_name: Optional[str] = None,
    special_keywords: Optional[list] = None,
) -> str: ...


@overload
def format_inlet(
    inlet: None,
    units: str = "m",
    key_name: Optional[str] = None,
    special_keywords: Optional[list] = None,
) -> None: ...


def format_inlet(
    inlet: Optional[str],
    units: str = "m",
    key_name: Optional[str] = None,
    special_keywords: Optional[list] = None,
) -> Optional[str]:
    """
    Make sure inlet / height name conforms to standard. The standard
    imposed can depend on the associated key_name itself (can
    be supplied as an option to check).

    This standard is as follows:
     - number followed by unit
     - number alone if unit / derviative is specified at the end of key_name (e.g. station_height_masl)
     - unchanged if this is one of the special keywords (by default "multiple" or "various")

    Other considerations:
     - For units of "m", we will also look for "magl" and "masl" (metres above ground and sea level)
     - If the input string just contains numbers, it is assumed this is already within the correct unit.

    Args:
        inlet: Inlet / Height value in the specified units
        units: Units for the inlet value ("m" by default)
        key_name: Name of the associated key. This is optional but will be used to
            determine whether the unit value should be added to the output string.
        special_keywords: Specify special keywords inlet could be set to
            If so do not apply any formatting.
            If this is not set a special keyword of "multiple" and "column" will still be allowed.
    Returns:
        str: formatted inlet string / None

    Usage:
        >>> format_inlet("10")
            "10m"
        >>> format_inlet("10m")
            "10m"
        >>> format_inlet("10magl")
            "10m"
        >>> format_inlet("10.111")
            "10.1m"
        >>> format_inlet("multiple")
            "multiple"
        >>> format_inlet("10m", key_name="inlet")
            "10m"
        >>> format_inlet("10m", key_name="inlet_magl")
            "10"
        >>> format_inlet("10m", key_name="station_height_masl")
            "10"
    """
    if inlet is None:
        return None

    # By default the special keyword is "multiple" for data containing multiple inlets.
    # This will be included if data is a combined object from the object store.
    if special_keywords is None:
        special_keywords = ["multiple", "column"]

    # Check if inlet is set to a special keyword
    if inlet in special_keywords:
        return inlet

    # Define set of options associated with units. For "m" this include
    # "magl" and "masl" (metres above ground and sea level).
    if units == "m":
        unit_options = ["m", "magl", "masl"]
    else:
        unit_options = [cast(str, units)]

    # Check whether unit is needed in string output.
    # This is dependent on whether the key name itself contains the unit value
    # (or derivative). If so, the unit itself is not needed in the value.
    unit_needed = True
    if key_name is not None:
        for value in unit_options:
            if key_name.split("_")[-1] == value:
                unit_needed = False

    # Check if input inlet just contains numbers and no unit
    # If so assume the units are metres and add this to the end of the string
    try:
        inlet_float = float(inlet)
    except ValueError:
        pass
    else:
        if inlet_float.is_integer():
            if unit_needed:
                inlet = f"{inlet_float:.0f}{units}"
            else:
                inlet = f"{inlet_float:.0f}"
        else:
            if unit_needed:
                inlet = f"{inlet_float:.1f}{units}"
            else:
                inlet = f"{inlet_float:.1f}"

        return str(inlet)

    # If we were unable to cast inlet as a float
    # check if inlet ends with unit or unit derivative
    # e.g. "magl" and "masl" would need to replaced with "m" or be removed
    for value in unit_options:
        if inlet.endswith(value):
            if unit_needed:
                inlet = inlet.replace(value, units)
            else:
                inlet = inlet.rstrip(value)
            break
    # else:
    #     raise ValueError(f"Did not recognise input for inlet: {inlet}")

    return str(inlet)


def extract_height_name(
    site: str,
    network: Optional[str] = None,
    inlet: Optional[str] = None,
    site_filepath: optionalPathType = None,
) -> Optional[Union[str, list]]:
    """
    Extract the relevant height associated with NAME from the
    "height_name" variable, if present from site_info data.

    This expects the "height_name" variable to be one of:
      - list containing the same number of items as inlets for the site
      - dictionary containing the mapping between inlets and heights
        used in NAME.

    Args:
        site : Site code
        network: Name of the associated network for the site
        inlet: Observation inlet / height value in the specified units
        site_filepath: Alternative site info file. Defaults to openghg_defs input.
    Returns:
        str : appropriate height name value extracted from site_info
        list: multiple height name options extracted from site_info
        None: if value not found or ambiguous.
    """
    from openghg.util import get_site_info

    site_data = get_site_info(site_filepath=site_filepath)

    if site:
        site_upper = site.upper()

    if network is None:
        network = next(iter(site_data[site_upper]))
    else:
        network = network.upper()

    height_name_attr = "height_name"
    height_attr = "height"

    if site_upper in site_data:
        site_metadata = site_data[site_upper][network]
        if height_name_attr in site_metadata:
            # Extract height_name variable from the site_metadata
            height_name_extracted = site_metadata[height_name_attr]
            # If this is a list, check and try and extract appropriate value.
            if isinstance(height_name_extracted, list):
                # Check if multiple values for height_name_extracted are present (list > 1)
                if len(height_name_extracted) == 1:
                    height_name: Optional[str] = height_name_extracted[0]
                else:
                    # If this is ambiguous, check "height" attr to match against site inlet value
                    # This assumes two lists of the same length map to each other with translating values
                    if (inlet is not None) and (height_attr in site_metadata):
                        height_values = site_metadata[height_attr]
                        if len(height_values) == len(height_name_extracted) and (inlet in height_values):
                            index = height_values.index(inlet)
                            height_name = height_name_extracted[index]
                        else:
                            logger.warning(
                                f"Ambiguous '{height_name_attr}' in site_info. "
                                f"Unable to extract from: height_name = {height_name_extracted} using height = {inlet}"
                            )
                            height_name = None
                    else:
                        logger.warning(
                            f"Ambiguous '{height_name_attr}' in site_info. "
                            f"Unable to extract from: height_name = {height_name_extracted}"
                        )
                        height_name = None
            elif isinstance(height_name_extracted, dict):
                if (inlet is not None) and (inlet in height_name_extracted):
                    height_name = height_name_extracted[inlet]
                else:
                    logger.warning(
                        f"Unable to interpret {height_name_extracted}. Please supply or check supplied inlet value: {inlet}"
                    )
                    height_name = None
        else:
            height_name = None
    else:
        height_name = None

    return height_name



from typing import List, Optional, Union, Dict, Any

from openghg.util import load_json
from openghg.types import optionalPathType

__all__ = ["get_species_info", "synonyms", "species_lifetime", "check_lifetime_monthly", "molar_mass"]


def get_species_info(species_filepath: optionalPathType = None) -> Dict[str, Any]:
    """Extract data from species info JSON file as a dictionary.

    This uses the data stored within openghg_defs/species_info JSON file by default.

    Args:
        species_filepath: Alternative species info file.
    Returns:
        dict: Data from species JSON file
    """
    from openghg_defs import species_info_file

    fpath = species_info_file if species_filepath is None else species_filepath

    return load_json(path=fpath)


def synonyms(
    species: str,
    lower: bool = True,
    allow_new_species: bool = True,
    species_filepath: optionalPathType = None,
) -> str:
    """Check to see if there are other names that we should be using for
    a particular input. E.g. If CFC-11 or CFC11 was input, go on to use cfc11.

    Args:
        species : Input string that you're trying to match
        lower : Return all lower case
        allow_new_species : Return original value (may be lower case)
            if this (or a synonym) is not found in the database.
            If False, raise a ValueError.
        species_filepath: Alternative species info file. Defaults to openghg_defs input.
    Returns:
        str: Matched species string

    TODO: Decide if we need to make this lower case or not.
    Included this here so this occurs in one place which can be linked to
    and changed if needed.
    """
    # If the species value is inert it should directly return rather than going through below logic
    if species.lower() == "inert":
        return species.lower()

    # Load in the species data
    species_data = get_species_info(species_filepath=species_filepath)

    # First test whether site matches keys (case insensitive)
    matched_strings = [k for k in species_data if k.upper() == species.upper()]

    # Used to access the alternative names in species_data
    alt_label = "alt"

    # If not found, search synonyms
    if not matched_strings:
        for key in species_data:
            # Iterate over the alternative labels and check for a match
            matched_strings = [s for s in species_data[key][alt_label] if s.upper() == species.upper()]

            if matched_strings:
                matched_strings = [key]
                break

    if matched_strings:
        updated_species = str(matched_strings[0])
        if lower:
            updated_species = updated_species.lower()
        return updated_species
    else:
        if not allow_new_species:
            raise ValueError(f"Unable to find species (or synonym) in database {species}")
        if lower:
            species = species.lower()
        return species


LifetimeType = Optional[Union[str, List[str]]]


def species_lifetime(species: Union[str, None], species_filepath: optionalPathType = None) -> LifetimeType:
    """Find species lifetime.
    This can either be labelled as "lifetime" or "lifetime_monthly".

    Note: no species synonyms accepted yet

    Args:
        species : Species name e.g. "ch4" or "co2"
        species_filepath: Alternative species info file. Defaults to openghg_defs input.
    Returns:
        str / list / None : Extracted lifetime or None is no lifetime was present.
    """
    species_data = get_species_info(species_filepath=species_filepath)

    if species is not None:
        species_label = synonyms(species, lower=False, allow_new_species=False)
        species_data = species_data[species_label]
    else:
        return None

    lifetime_keywords = ["lifetime", "lifetime_monthly"]
    for key in lifetime_keywords:
        try:
            lifetime: Optional[list] = species_data[key]
        except KeyError:
            continue
        else:
            break
    else:
        lifetime = None

    return lifetime


def check_lifetime_monthly(lifetime: LifetimeType) -> bool:
    """Check whether retrieved lifetime value represents monthly lifetimes.
    This checks whether lifetime is a list and contains 12 values.

    Args:
        lifetime : str or list representation of lifetime value
    Returns:
        bool : True of lifetime matches criteria for monthly data, False otherwise
    Raises ValueError:
        if lifetime is a list but does not contain exactly 12 entries, one for each month
    """
    if isinstance(lifetime, list):
        if len(lifetime) == 12:
            return True
        else:
            raise ValueError(f"Invalid input for lifetime: {lifetime}")
    else:
        return False


def molar_mass(species: str, species_filepath: optionalPathType = None) -> float:
    """Extracts the molar mass of a species.

    Args:
        species : Species name
        species_filepath: Alternative species info file. Defaults to openghg_defs input.
    Returns:
        float : Molar mass of species
    """
    species_data = get_species_info(species_filepath=species_filepath)

    species_label = synonyms(species, lower=False, allow_new_species=False)
    molmass = float(species_data[species_label]["mol_mass"])

    return molmass



from pathlib import Path
from typing import Dict, List, Optional, Union, Any
from pandas import Timedelta
import warnings

from openghg.cloud import create_file_package, create_post_dict
from openghg.objectstore import get_writable_bucket
from openghg.util import running_on_hub
from openghg.types import optionalPathType, multiPathType
from numcodecs import Blosc
import logging

logger = logging.getLogger("openghg.standardise")


def standardise(data_type: str, filepath: multiPathType, store: Optional[str] = None, **kwargs: Any) -> Dict:
    """Generic standardise function, used by data-type specific versions.

    Args:
        data_type: type of data to standardise
        filepath: path to file(s) to standardise
        store: Name of object store to write to, required if user has access to more than one
        writable store
        **kwargs: data type specific arguments, see specific implementations below.
    Returns:
        dict: Dictionary of result data.
    """
    from openghg.store import get_data_class

    dclass = get_data_class(data_type)
    bucket = get_writable_bucket(name=store)

    compression = kwargs.get("compression", True)
    compressor = kwargs.get("compressor")

    if compression:
        if compressor is None:
            compressor = Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)
    else:
        logger.info("Compression disabled")
        compressor = None

    kwargs["compressor"] = compressor

    try:
        del kwargs["compression"]
    except KeyError:
        pass

    with dclass(bucket=bucket) as dc:
        result = dc.read_file(filepath=filepath, **kwargs)
    return result


def standardise_surface(
    source_format: str,
    network: str,
    site: str,
    filepath: Optional[multiPathType] = None,
    filepaths: Optional[multiPathType] = None,
    inlet: Optional[str] = None,
    height: Optional[str] = None,
    instrument: Optional[str] = None,
    sampling_period: Optional[Union[Timedelta, str]] = None,
    calibration_scale: Optional[str] = None,
    measurement_type: str = "insitu",
    verify_site_code: bool = True,
    site_filepath: optionalPathType = None,
    store: Optional[str] = None,
    update_mismatch: str = "never",
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    force: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    chunks: Optional[Dict] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Standardise surface measurements and store the data in the object store.

    Args:
        filepath: Filepath(s)
        source_format: Data format, for example CRDS, GCWERKS
        site: Site code/name
        network: Network name
        inlet: Inlet height. Format 'NUMUNIT' e.g. "10m".
            If retrieve multiple files pass None, OpenGHG will attempt to
            extract this from the file.
        height: Alias for inlet.
        instrument: Instrument name
        sampling_period: Sampling period as pandas time code, e.g. 1m for 1 minute, 1h for 1 hour
        calibration_scale: Calibration scale for data
        measurement_type: Type of measurement e.g. insitu, flask
        verify_site_code: Verify the site code
        site_filepath: Alternative site info file (see openghg/openghg_defs repository for format).
            Otherwise will use the data stored within openghg_defs/data/site_info JSON file by default.
        store: Name of object store to write to, required if user has access to more than one
            writable store
        update_mismatch: This determines how mismatches between the internal data
            "attributes" and the supplied / derived "metadata" are handled.
            This includes the options:
                - "never" - don't update mismatches and raise an AttrMismatchError
                - "from_source" / "attributes" - update mismatches based on input attributes
                - "from_definition" / "metadata" - update mismatches based on input metadata
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
             - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
             - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
             - "n" / "no" - Allow current data to updated / deleted
        overwrite: Deprecated. This will use options for if_exists="new".
        force: Force adding of data even if this is identical to data stored.
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
            To disable chunking pass an empty dictionary.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    Returns:
        dict: Dictionary of result data
    """
    from openghg.cloud import call_function

    if filepath is None and filepaths is None:
        raise ValueError("One of `filepath` and `filepaths` must be specified.")
    elif filepath is None:
        filepath = filepaths
        warnings.warn(
            "The argument 'filepaths' will be deprecated in a future release. Please use 'filepath' instead.",
            FutureWarning,
        )

    if not isinstance(filepath, list):
        filepath = [filepath]

    if running_on_hub():
        # TODO: Use input for site_filepath here? How to include this?

        # To convert bytes to megabytes
        MB = 1e6
        # The largest file we'll just directly POST to the standardisation
        # function will be this big (megabytes)
        post_limit = 40  # MB

        metadata = {}
        metadata["site"] = site
        metadata["source_format"] = source_format
        metadata["network"] = network
        metadata["data_type"] = "surface"

        if inlet is not None:
            metadata["inlet"] = inlet
        if instrument is not None:
            metadata["instrument"] = instrument
        if sampling_period is not None:
            metadata["sampling_period"] = sampling_period

        responses = {}
        for fpath in filepath:
            gcwerks = False
            if source_format.lower() in ("gc", "gcwerks"):
                metadata["source_format"] = "gcwerks"

                try:
                    filepath = Path(fpath[0])
                except TypeError:
                    raise TypeError("We require both data and precision files for GCWERKS data.")
                else:
                    gcwerks = True
            else:
                filepath = Path(fpath)

            compressed_data, file_metadata = create_file_package(filepath=filepath, obs_type="surface")
            compressed_size = len(compressed_data) / MB

            if compressed_size > post_limit:
                raise NotImplementedError("Compressed size over 40 MB, not currently supported.")

            to_post = {
                "function": "standardise",
                "data": compressed_data,
                "metadata": metadata,
                "file_metadata": file_metadata,
            }

            if gcwerks:
                precision_filepath = Path(fpath[1])
                compressed_prec, prec_file_metadata = create_file_package(
                    filepath=precision_filepath, obs_type="surface"
                )

                to_post["precision_data"] = compressed_prec
                to_post["precision_file_metadata"] = prec_file_metadata

            # else:
            # If we want chunked uploading what do we do?
            # raise NotImplementedError
            # tmp_dir = tempfile.TemporaryDirectory()
            # compressed_filepath = Path(tmp_dir.name).joinpath(f"{filepath.name}.tar.gz")
            # # Compress in place and then upload
            # with tarfile.open(compressed_filepath, mode="w:gz") as tar:
            #     tar.add(filepath)
            # compressed_data = compressed_filepath.read_bytes()

            fn_response = call_function(data=to_post)

            responses[filepath.name] = fn_response["content"]

        return responses
    else:
        return standardise(
            store=store,
            data_type="surface",
            filepath=filepath,
            source_format=source_format,
            network=network,
            site=site,
            inlet=inlet,
            height=height,
            instrument=instrument,
            sampling_period=sampling_period,
            calibration_scale=calibration_scale,
            measurement_type=measurement_type,
            overwrite=overwrite,
            verify_site_code=verify_site_code,
            site_filepath=site_filepath,
            update_mismatch=update_mismatch,
            if_exists=if_exists,
            save_current=save_current,
            force=force,
            compression=compression,
            compressor=compressor,
            filters=filters,
            chunks=chunks,
            optional_metadata=optional_metadata,
        )


def standardise_column(
    filepath: Union[str, Path],
    satellite: Optional[str] = None,
    domain: Optional[str] = None,
    selection: Optional[str] = None,
    site: Optional[str] = None,
    species: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    platform: str = "satellite",
    source_format: str = "openghg",
    store: Optional[str] = None,
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    force: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    chunks: Optional[Dict] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Read column observation file

    Args:
        filepath: Path of observation file
        satellite: Name of satellite (if relevant)
        domain: For satellite only. If data has been selected on an area include the
            identifier name for domain covered. This can map to previously defined domains
            (see openghg_defs "domain_info.json" file) or a newly defined domain.
        selection: For satellite only, identifier for any data selection which has been
            performed on satellite data. This can be based on any form of filtering, binning etc.
            but should be unique compared to other selections made e.g. "land", "glint", "upperlimit".
            If not specified, domain will be used.
        site : Site code/name (if relevant). Can include satellite OR site.
        species: Species name or synonym e.g. "ch4"
        instrument: Instrument name e.g. "TANSO-FTS"
        network: Name of in-situ or satellite network e.g. "TCCON", "GOSAT"
        platform: Type of platform. Should be one of:
            - "satellite"
            - "site"
        source_format : Type of data being input e.g. openghg (internal format)
        store: Name of store to write to
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
             - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
             - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
             - "n" / "no" - Allow current data to updated / deleted
        overwrite: Deprecated. This will use options for if_exists="new".
        force: Force adding of data even if this is identical to data stored.
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking
            To disable chunking pass an empty dictionary.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    Returns:
        dict: Dictionary containing confirmation of standardisation process.
    """
    from openghg.cloud import call_function

    filepath = Path(filepath)

    if running_on_hub():
        compressed_data, file_metadata = create_file_package(filepath=filepath, obs_type="footprints")

        metadata = {
            "site": site,
            "satellite": satellite,
            "domain": domain,
            "selection": selection,
            "site": site,
            "species": species,
            "network": network,
            "instrument": instrument,
            "platform": platform,
            "source_format": source_format,
            "overwrite": overwrite,
        }

        metadata = {k: v for k, v in metadata.items() if v is not None}

        to_post = create_post_dict(
            function_name="standardise", data=compressed_data, metadata=metadata, file_metadata=file_metadata
        )

        fn_response = call_function(data=to_post)
        response_content: Dict = fn_response["content"]
        return response_content
    else:
        return standardise(
            store=store,
            data_type="column",
            filepath=filepath,
            satellite=satellite,
            domain=domain,
            selection=selection,
            site=site,
            species=species,
            network=network,
            instrument=instrument,
            platform=platform,
            source_format=source_format,
            overwrite=overwrite,
            if_exists=if_exists,
            save_current=save_current,
            force=force,
            compression=compression,
            compressor=compressor,
            filters=filters,
            chunks=chunks,
            optional_metadata=optional_metadata,
        )


def standardise_bc(
    filepath: Union[str, Path],
    species: str,
    bc_input: str,
    domain: str,
    period: Optional[Union[str, tuple]] = None,
    continuous: bool = True,
    store: Optional[str] = None,
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    force: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    chunks: Optional[Dict] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Standardise boundary condition data and store it in the object store.

    Args:
        filepath: Path of boundary conditions file
        species: Species name
        bc_input: Input used to create boundary conditions. For example:
            - a model name such as "MOZART" or "CAMS"
            - a description such as "UniformAGAGE" (uniform values based on AGAGE average)
        domain: Region for boundary conditions
        period: Period of measurements, if not passed this is inferred from the time coords
        continuous: Whether time stamps have to be continuous.
        store: Name of store to write to
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
             - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
             - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
             - "n" / "no" - Allow current data to updated / deleted
        overwrite: Deprecated. This will use options for if_exists="new".
        force: Force adding of data even if this is identical to data stored.
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking
            To disable chunking pass an empty dictionary.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    returns:
        dict: Dictionary containing confirmation of standardisation process.
    """
    from openghg.cloud import call_function

    filepath = Path(filepath)

    if running_on_hub():
        compressed_data, file_metadata = create_file_package(filepath=filepath, obs_type="bc")

        metadata = {
            "species": species,
            "bc_input": bc_input,
            "domain": domain,
            "continuous": continuous,
            "overwrite": overwrite,
        }

        if period is not None:
            metadata["period"] = period

        to_post = create_post_dict(
            function_name="standardise", data=compressed_data, metadata=metadata, file_metadata=file_metadata
        )

        fn_response = call_function(data=to_post)
        response_content: Dict = fn_response["content"]
        return response_content
    else:
        return standardise(
            store=store,
            data_type="boundary_conditions",
            filepath=filepath,
            species=species,
            bc_input=bc_input,
            domain=domain,
            period=period,
            continuous=continuous,
            overwrite=overwrite,
            if_exists=if_exists,
            save_current=save_current,
            force=force,
            compression=compression,
            compressor=compressor,
            filters=filters,
            chunks=chunks,
            optional_metadata=optional_metadata,
        )


def standardise_footprint(
    filepath: Union[str, Path, List],
    site: str,
    domain: str,
    model: str,
    inlet: Optional[str] = None,
    height: Optional[str] = None,
    met_model: Optional[str] = None,
    species: Optional[str] = None,
    network: Optional[str] = None,
    source_format: str = "acrg_org",
    period: Optional[Union[str, tuple]] = None,
    chunks: Optional[Dict] = None,
    continuous: bool = True,
    retrieve_met: bool = False,
    store: Optional[str] = None,
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    force: bool = False,
    high_spatial_resolution: bool = False,
    time_resolved: bool = False,
    high_time_resolution: bool = False,
    short_lifetime: bool = False,
    sort: bool = False,
    drop_duplicates: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Reads footprint data files and returns the UUIDs of the Datasources
    the processed data has been assigned to

    Args:
        filepath: Path(s) of file to standardise
        site: Site name
        domain: Domain of footprints
        model: Model used to create footprint (e.g. NAME or FLEXPART)
        inlet: Height above ground level in metres. Format 'NUMUNIT' e.g. "10m"
        height: Alias for inlet. One of height or inlet must be included.
        met_model: Underlying meteorlogical model used (e.g. UKV)
        species: Species name. Only needed if footprint is for a specific species e.g. co2 (and not inert)
        network: Network name
        source_format: Format of the input data format, for example acrg_org
        period: Period of measurements. Only needed if this can not be inferred from the time coords
        chunks: Chunk schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking
            by OpenGHG as per the TODO RELEASE: add link to documentation. To disable chunking pass an empty dictionary.
        continuous: Whether time stamps have to be continuous.
        retrieve_met: Whether to also download meterological data for this footprints area
        high_spatial_resolution : Indicate footprints include both a low and high spatial resolution.
        time_resolved: Indicate footprints are high time resolution (include H_back dimension)
            Note this will be set to True automatically for Carbon Dioxide data.
        short_lifetime: Indicate footprint is for a short-lived species. Needs species input.
            Note this will be set to True if species has an associated lifetime.
        high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
        store: Name of store to write to
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
             - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
             - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
             - "n" / "no" - Allow current data to updated / deleted        overwrite: Deprecated. This will use options for if_exists="new".
        force: Force adding of data even if this is identical to data stored.
        sort: Sort data in by time
        drop_duplicates: Drop duplicate timestamps, keeping the first value
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    Returns:
        dict / None: Dictionary containing confirmation of standardisation process. None
        if file already processed.
    """
    from openghg.cloud import call_function

    if high_time_resolution:
        warnings.warn(
            "This argument is deprecated and will be replaced in future versions with time_resolved.",
            DeprecationWarning,
        )
        time_resolved = high_time_resolution

    if running_on_hub():
        raise NotImplementedError("Cloud support not yet implemented.")
        compressed_data, file_metadata = create_file_package(filepath=filepath, obs_type="footprints")

        metadata = {
            "site": site,
            "domain": domain,
            "model": model,
            "inlet": inlet,
            "height": height,
            "continuous": continuous,
            "retrieve_met": retrieve_met,
            "high_spatial_resolution": high_spatial_resolution,
            "time_resolved": time_resolved,
            "overwrite": overwrite,
            "met_model": met_model,
            "species": species,
            "network": network,
            "period": period,
            "chunks": chunks,
        }

        metadata = {k: v for k, v in metadata.items() if v is not None}

        to_post = create_post_dict(
            function_name="standardise", data=compressed_data, metadata=metadata, file_metadata=file_metadata
        )

        fn_response = call_function(data=to_post)
        response_content: Dict = fn_response["content"]
        return response_content
    else:
        return standardise(
            store=store,
            data_type="footprints",
            filepath=filepath,
            site=site,
            domain=domain,
            model=model,
            inlet=inlet,
            height=height,
            met_model=met_model,
            species=species,
            network=network,
            source_format=source_format,
            period=period,
            chunks=chunks,
            continuous=continuous,
            retrieve_met=retrieve_met,
            high_spatial_resolution=high_spatial_resolution,
            time_resolved=time_resolved,
            short_lifetime=short_lifetime,
            overwrite=overwrite,
            if_exists=if_exists,
            save_current=save_current,
            force=force,
            compression=compression,
            compressor=compressor,
            filters=filters,
            sort=sort,
            drop_duplicates=drop_duplicates,
            optional_metadata=optional_metadata,
        )


def standardise_flux(
    filepath: Union[str, Path],
    species: str,
    source: str,
    domain: str,
    database: Optional[str] = None,
    source_format: str = "openghg",
    database_version: Optional[str] = None,
    model: Optional[str] = None,
    time_resolved: bool = False,
    high_time_resolution: bool = False,
    period: Optional[Union[str, tuple]] = None,
    chunks: Optional[Dict] = None,
    continuous: bool = True,
    store: Optional[str] = None,
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    force: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Process flux / emissions data

    Args:
        filepath: Path of flux / emissions file
        species: Species name
        source: Flux / Emissions source
        domain: Flux / Emissions domain
        source_format: Data format, for example openghg, intem
        date : Date as a string e.g. "2012" or "201206" associated with emissions as a string.
               Only needed if this can not be inferred from the time coords
        time_resolved: If this is a high resolution file
        high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
        period: Period of measurements, if not passed this is inferred from the time coords
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
            To disable chunking pass an empty dictionary.
        continuous: Whether time stamps have to be continuous.
        store: Name of store to write to
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
             - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
             - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
             - "n" / "no" - Allow current data to updated / deleted
        overwrite: Deprecated. This will use options for if_exists="new".
        force: Force adding of data even if this is identical to data stored.
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    returns:
        dict: Dictionary of Datasource UUIDs data assigned to
    """
    from openghg.cloud import call_function

    filepath = Path(filepath)

    if high_time_resolution:
        warnings.warn(
            "This argument is deprecated and will be replaced in future versions with time_resolved.",
            DeprecationWarning,
        )
        time_resolved = high_time_resolution

    if running_on_hub():
        compressed_data, file_metadata = create_file_package(filepath=filepath, obs_type="flux")

        metadata = {
            "species": species,
            "source": source,
            "domain": domain,
            "time_resolved": time_resolved,
            "continuous": continuous,
            "overwrite": overwrite,
            "chunks": chunks,
            "period": period,
            "source_format": source_format,
        }

        optional_keywords = {"database": database, "database_version": database_version, "model": model}
        for key, value in optional_keywords.items():
            if value is not None:
                metadata[key] = value

        metadata = {k: v for k, v in metadata.items()}

        to_post = create_post_dict(
            function_name="standardise", data=compressed_data, metadata=metadata, file_metadata=file_metadata
        )

        fn_response = call_function(data=to_post)
        response_content: Dict = fn_response["content"]
        return response_content
    else:
        return standardise(
            data_type="flux",
            store=store,
            filepath=filepath,
            species=species,
            source=source,
            domain=domain,
            database=database,
            database_version=database_version,
            model=model,
            time_resolved=time_resolved,
            period=period,
            continuous=continuous,
            chunks=chunks,
            overwrite=overwrite,
            if_exists=if_exists,
            save_current=save_current,
            force=force,
            compression=compression,
            compressor=compressor,
            filters=filters,
            optional_metadata=optional_metadata,
        )


def standardise_eulerian(
    filepath: Union[str, Path],
    model: str,
    species: str,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    setup: Optional[str] = None,
    if_exists: str = "auto",
    save_current: str = "auto",
    overwrite: bool = False,
    store: Optional[str] = None,
    force: bool = False,
    compression: bool = True,
    compressor: Optional[Any] = None,
    filters: Optional[Any] = None,
    chunks: Optional[Dict] = None,
    optional_metadata: Optional[Dict] = None,
) -> Dict:
    """Read Eulerian model output

    Args:
        filepath: Path of Eulerian model species output
        model: Eulerian model name
        species: Species name
        start_date: Start date (inclusive) associated with model run
        end_date: End date (exclusive) associated with model run
        setup: Additional setup details for run
        if_exists: What to do if existing data is present.
            - "auto" - checks new and current data for timeseries overlap
                - adds data if no overlap
                - raises DataOverlapError if there is an overlap
            - "new" - just include new data and ignore previous
            - "combine" - replace and insert new data into current timeseries
        save_current: Whether to save data in current form and create a new version.
            - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
            - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
            - "n" / "no" - Allow current data to updated / deleted
        overwrite: Deprecated. This will use options for if_exists="new".
        store: Name of object store to write to, required if user has access to more than one
        writable store
        force: Force adding of data even if this is identical to data stored.
        compression: Enable compression in the store
        compressor: A custom compressor to use. If None, this will default to
            `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
            See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
        filters: Filters to apply to the data on storage, this defaults to no filtering. See
            https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
            To disable chunking pass an empty dictionary.
        optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
    Returns:
        dict: Dictionary of result data
    """
    if running_on_hub():
        raise NotImplementedError("Serverless not implemented yet for Eulerian model.")
    else:
        return standardise(
            store=store,
            data_type="eulerian_model",
            filepath=filepath,
            model=model,
            species=species,
            start_date=start_date,
            end_date=end_date,
            setup=setup,
            overwrite=overwrite,
            if_exists=if_exists,
            force=force,
            save_current=save_current,
            compression=compression,
            compressor=compressor,
            filters=filters,
            chunks=chunks,
            optional_metadata=optional_metadata,
        )


def standardise_from_binary_data(
    store: str,
    data_type: str,
    binary_data: bytes,
    metadata: dict,
    file_metadata: dict,
    **kwargs: Any,
) -> Optional[Dict]:
    """Standardise binary data from serverless function.
        The data dictionary should contain sub-dictionaries that contain
        data and metadata keys.

    args:
        store: Name of object store to write to, required if user has access to more than one
        writable store
        data_type: type of data to standardise
        binary_data: Binary measurement data
        metadata: Metadata
        file_metadata: File metadata such as original filename
        **kwargs: data type specific arguments, see specific implementations in data classes.
    returns:
        Dictionary of result data.
    """
    from openghg.store import get_data_class

    dclass = get_data_class(data_type)
    bucket = get_writable_bucket(name=store)

    with dclass(bucket) as dc:
        result = dc.read_data(
            binary_data=binary_data, metadata=metadata, file_metadata=file_metadata, **kwargs
        )
    return result
